

/* Module: Wavelets.
	Please note that this module is highly experimental.


*/

/*		MODULE FOR CRYSTAL IMAGE 0.85 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */




#include <gtk/gtk.h>
#include <gdk/gdk.h>

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>



#include "includeall.h"









/***************************************************************************************************************

  ***    *   *   ****    ****      *     *   *   ****            *****   *       *****
 *   *   *   *   *   *   *   *    * *    *   *   *   *           *       *         *
 *       *   *   *   *   *   *   *   *   **  *   *   *           *       *         *
  ***    *   *   ****    ****    *   *   * * *   *   *           ****    *         *
     *   *   *   *   *   *   *   *****   *  **   *   *           *       *         *
 *   *   *   *   *   *   *   *   *   *   *   *   *   *           *       *         *
  ***     ***    ****    ****    *   *   *   *   ****            *       *****     *

*/

/* The subband filters perform one step of DWT decomposision or synthesis.
	These subband filters operate on an 1D array of float data (length
	must be a power of two or - less strictly - an even number) and return
	the scaling filter coefficients in the lower half of the array and the 
	wavelet coefficients in the upper half of the array.
	The variable idwt determines whether we want a forward DWT (idwt==0)
	or an inverse DWT (idwt!=0).

	In this experimental module, we have only two standalone subband filters,
	the Daubechies-4 and the Haar filter. All others (including these two, but
	in a generalized fashion) are in the wavelet core module.
*/





/*	Daubechies-4 one-stage subband coding filter
		This filter takes a one-dimensional array of data [0...n-1], length n MUST be a power of two
		and performs one stage of subband coding (idwt=0 -> forward transform).  The output is interleaved,
		let's say the input is l0, l1, l2... (for Lambda), and the output is the next level's l0, l1, ...
		from the scaling filter and g0, g1, g2... from the wavelet filter. Then, the output has the
		sequence l0, g0, l1, g1, l2, g2... - However, the formulation here (two lines in the FOR loop)
		provides the data shuffling as well, i.e. the output is l0, l1, l2, ... , g0, g1, g2, ...
		whereby the sequence g0, g1, g2... starts at nh=n/2.
		
		Conversely, the inverse DWT (variable idwt=1) requires the sequence l0, l1, l2, ... , g0, g1, g2, ...
		as input and joins them to form the previous stage's l0, l1, l2, l3, ...
		
		Note that the length of the data array may be longer than n - in that case, the values above n-1
		remain untouched. This is important once one subband stange has executed and we have an array of
		l0, l1, l2, ... , g0, g1, g2, ... (i.e. reshuffled) - then we divide n by two and call
		the subband coding function again.
*/



void daub4_subband (float *data, unsigned long n, DWTDIRECTION idwt, wavefilt dummy)
{

float *buf;
unsigned long nh,nh1,i,j;
float cdaub4[4];
float fsqrt2, sqrt3;

	if (n < 4) return;
	
	buf= (float *) calloc (n+1, sizeof (float));		/* temporary storage buffer */
	nh = n >> 1;
	nh1 = (nh)+1;

	/* Use the explicit equations for the Daub-4 filter coefficients to obtain full precision 
		even if we'd decide to go for double rather than float. Note that this is an inefficiency
		as the coefficients get recomputed  every time we use this subband filter. */
	
	fsqrt2 = 4.0*sqrt(2.0); sqrt3 = sqrt(3.0);
	cdaub4[0] = (1+sqrt3)/fsqrt2;
	cdaub4[1] = (3+sqrt3)/fsqrt2;
	cdaub4[2] = (3-sqrt3)/fsqrt2;
	cdaub4[3] = (1-sqrt3)/fsqrt2;

	if (idwt==W_FORWARD) 			/* Forward subband coding filter (forward one-level DWT). */
	{
		i=0;
		for (j=0; j<n-3; j+=2)
		{
			buf[i]    = cdaub4[0]*data[j] + cdaub4[1]*data[j+1] + cdaub4[2]*data[j+2] + cdaub4[3]*data[j+3];
			buf[i+nh] = cdaub4[3]*data[j] - cdaub4[2]*data[j+1] + cdaub4[1]*data[j+2] - cdaub4[0]*data[j+3];
			i++;
		}
		
		buf[i]    = cdaub4[0]*data[n-2] + cdaub4[1]*data[n-1] + cdaub4[2]*data[0] + cdaub4[3]*data[1];
		buf[i+nh] = cdaub4[3]*data[n-2] - cdaub4[2]*data[n-1] + cdaub4[1]*data[0] - cdaub4[0]*data[1];

	} 
	else 			/* Inverse filter (inverse DWT). */
	{
		buf[0] = cdaub4[2]*data[nh-1] + cdaub4[1]*data[n-1] + cdaub4[0]*data[0] + cdaub4[3]*data[nh];
		buf[1] = cdaub4[3]*data[nh-1] - cdaub4[0]*data[n-1] + cdaub4[1]*data[0] - cdaub4[2]*data[nh];
		j=2;
		
		for (i=0; i<nh-1; i++) 
		{

			buf[j++] = cdaub4[2]*data[i] + cdaub4[1]*data[i+nh] + cdaub4[0]*data[i+1] + cdaub4[3]*data[i+nh1];
			buf[j++] = cdaub4[3]*data[i] - cdaub4[0]*data[i+nh] + cdaub4[1]*data[i+1] - cdaub4[2]*data[i+nh1];
		}
	}
	for (i=0; i<n; i++) data[i]=buf[i];
	free (buf);

}


/* This one may still have issues, particularly with the high end. Matrix wraparound? */

void haar_subband (float *data, unsigned long n, DWTDIRECTION idwt, wavefilt dummy)
{

float *buf;
unsigned long nh,i,j;
float chaar[4];

	if (n < 4) return;
	
	chaar[0] = 1/sqrt(2);
	chaar[1] = chaar[0];
	
	buf= (float *) calloc (n+1, sizeof (float));		/* temporary storage buffer */
	nh = n >> 1;
	


	if (idwt==W_FORWARD) 			/* Forward subband coding filter (forward one-level DWT). */
	{
		i=0;
		for (j=0; j<n-1; j+=2)
		{
			buf[i]    = chaar[0]*data[j] + chaar[1]*data[j+1];
			buf[i+nh] = chaar[1]*data[j] - chaar[0]*data[j+1];
			i++;
		}
		
	} 
	else 			/* Inverse filter (inverse DWT). */
	{
		j=0;
		for (i=0; i<nh-1; i++) 
		{

			buf[j++] = chaar[0]*data[i] + chaar[1]*data[i+nh];
			buf[j++] = chaar[1]*data[i] - chaar[0]*data[i+nh];
		}
	}
	for (i=0; i<n; i++) data[i]=buf[i];
	free (buf);

}











/*
 *****   *****    ***    *****    ***
   *     *       *   *     *     *   *
   *     *       *         *     *
   *     ****     ***      *      ***
   *     *           *     *         *
   *     *       *   *     *     *   *
   *     *****    ***      *      ***

*/

/* A test function. Generate a test signal and decompose. Display
   transform and reconstruct, display again.

   The test signal is known as "Manhattan Skyline" and constains steps, ramps, and
   other discontinuities. The Manhattan skyline is 2048 points long.
*/


void manhattan_skyline (float *x, float *y)
{
int i;

	for (i=0; i<128; i++) 		y[i] = 2.0;
	for (i=128; i<160; i++) 	y[i] = 5.0;
	for (i=160; i<320; i++) 	y[i] = 1.0;
	for (i=320; i<336; i++) 	y[i] = 0.0;
	for (i=336; i<512; i++) 	y[i] = 3.0;
	for (i=512; i<544; i++) 	y[i] = 7.0;
	for (i=544; i<800; i++) 	y[i] = 0.0;
	for (i=800; i<960; i++) 	y[i] = (i-800.0)/160.0;
	for (i=960; i<1024; i++) 	y[i] = 1.0;
	for (i=1024; i<1200; i++) 	y[i] = 3.0;
	for (i=1200; i<1360; i++) 	y[i] = 2.0;
	for (i=1360; i<1600; i++) 	y[i] = (1600.0-i)/240.0;
	for (i=1600; i<1616; i++) 	y[i] = 0.0;
	for (i=1616; i<1632; i++) 	y[i] = 10.0;
	for (i=1632; i<1760; i++) 	y[i] = (i-1632.0)/128.0;
	for (i=1760; i<2048; i++) 	y[i] = 1.0;

	for (i=0; i<2048; i++) x[i]=i;

}



void test_wavelet (int testfn, float tmult, int wfunc, int wshrinkfunc, wavefilt wfilt, void (*wt_subband)(float *, unsigned long, DWTDIRECTION, wavefilt) )
{
float *xdata, *ydata;
float stdev, buf;
int i,N;


	N = 2048;
	xdata = calloc (2048, sizeof (float));
	ydata = calloc (3*2048, sizeof (float));
	manhattan_skyline (xdata, ydata);
	stdev = 0.1;

	for (i=0; i<N; i++)			/* Add Gaussian noise to the signal */
	{
		ydata[i] += gauss_rnd()*stdev;
	}

	for (i=0; i<N; i++) ydata[i+N+N]=ydata[i];		/* Copy two up to keep original signal in Set 3 */

	if (testfn==0)										/* Low-level DWT with direct call to the subband functions */
	{
		dp (1, "1D Wavelet test function 0: Four-level decomposition and reconstruction\n");
		for(i=N; i>=128; i >>= 1)
			(*wt_subband)(ydata,i,W_FORWARD, wfilt);			/* four-level DWT decomposition */
		for (i=0; i<N; i++) ydata[i+N]=ydata[i];				/* Copy up into Set 2 to keep DWT */
		for(i=128; i<=N; i <<= 1)
			(*wt_subband)(ydata,i,W_INVERSE, wfilt);			/* Four-level IDWT */
	}
	else if (testfn==1)									/* High-level DWT through function dwt_1d */
	{
		dp (1, "1D Wavelet test function 1: N-level decomposition and reconstruction\n");
		dwt_1d (ydata, N, W_FORWARD, wfilt, (*wt_subband));			/* Wavelet transform */
		for (i=0; i<N; i++) ydata[i+N]=ydata[i];				/* Copy up to keep DWT */
		dwt_1d (ydata, N, W_INVERSE, wfilt, (*wt_subband));			/* Inverse Wavelet transform */
	}
	else if (testfn==2)									/* 1D denoising function */
	{
		dp (1, "1D Wavelet test function 2: Wavelet denoising\n");
		// void wvlt_1D_denoising  (float *data, int len, int wfunc, int wshrinkfunc, float tmult)
		wvlt_1D_denoising (ydata, N, wfunc, wshrinkfunc, tmult); 
		for (i=0; i<N; i++) 
		{
			buf = ydata[i];
			ydata[i+N] = ydata[i+N+N] - buf;				/* Green: Noise that was removed */
			ydata[i] = ydata[i+N+N];						/* Red: Original (noisy) signal */
			ydata[i+N+N] = buf;								/* Blue, drawn over red, denoised signal */
		}
	}

	multi_graphme (xdata, ydata, N, 3);				/* Red: Reconstructed. Green: DWT, Blue: Original */

	free (xdata); free (ydata);
}



/*+**********************************+******************+

  ####    #    #   #*#
 #    #   #    #    #
 #        #    #    #
 #   ##   #    #    #
 #    #   #    #    #
 #    #   #    #    #
  ####     ####    #*#

*/


void (*subband)(float *, unsigned long, DWTDIRECTION, wavefilt);			/* Pointer to the selected subband filter function */
wavefilt wfilt;	
GtkWidget *wvlt_funcmenu;
GtkWidget *wvlt_shrinkmenu;
GtkWidget *wvlt_typemenu;
GtkWidget *wvlt_label4, *wvlt_label1, *wvlt_label5;
GtkWidget *wvlt_tmult;

DWTDIRECTION idwt;
int wfunction, woperation, wshrinkfunc;
GtkAdjustment *wvlt_decomp_adj, *wvlt_tmult_adj;



/* We have a similar function in the core, but that function does not know the experimental
	standalone subband filters for Daub4 and Haar */

WAVELET_TYPE set_subband_ext (int wavelet)
{
WAVELET_TYPE wtype;
	
	wtype = W_NULL;
	switch (wavelet)
	{
		case 0:						/* Haar */
				wtype=W_HAAR;
				subband = &general_subband;
				break;
		case 1:						/* DAUB4 */
				wtype=W_DAUB4;
				subband = &general_subband;
				break;
		case 2:						/* DAUB8 */
				wtype=W_DAUB8;
				subband = &general_subband;
				break;
		case 3:						/* DAUB12 */
				wtype=W_DAUB12;
				subband = &general_subband;
				break;
		case 4:						/* DAUB20 */
				wtype=W_DAUB20;
				subband = &general_subband;
				break;
		case 5:						/* Coiflet-6 */
				wtype=W_COIF6;
				subband = &general_subband;
				break;
		case 6:						/* Coiflet-12 */
				wtype=W_COIF12;
				subband = &general_subband;
				break;
		case 7:						/* Symlet 4 */
				wtype=W_SYM4;
				subband = &general_subband;
				break;
		case 8:						/* Symlet 8 */
				wtype=W_SYM8;
				subband = &general_subband;
				break;
		case 9:						/* Symlet12 */
				wtype=W_SYM12;
				subband = &general_subband;
				break;
		case 10:						/* Standalone  Haar */
				subband = &haar_subband;
				return W_NULL;
				break;
		case 11:						/* Standalone DAUB4 */
				subband = &daub4_subband;
				return W_NULL;
				break;
		
		default:
				subband = &dead_subband;
				return 0;
	}
	return wtype;
}


void on_wvlt_ok_clicked  (GtkButton *button, gpointer user_data)
{
int decomp_level;
WAVELET_TYPE wtype;
float tmult;

	undo_nexus (&mainimg);

	decomp_level = wvlt_decomp_adj->value;
	tmult = wvlt_tmult_adj->value;

	if ((woperation==0) || (woperation==1))						/* DWT and IDWT */
	{
		wtype = set_subband_ext (wfunction);
		if (wtype!=W_NULL) wfilt = wavelet_prepare (wtype);
		dwt_2d (&mainimg, idwt, decomp_level, wfilt, *subband);
	}
	else if (woperation==2)										/* Wavelet shrinkage */
	{
		wavelet_shrink (&mainimg, decomp_level, wshrinkfunc, tmult);
	}
	else if (woperation==3)										/* All-in-one denoising */
	{
		wtype = set_subband_ext (wfunction);
		if (wtype!=W_NULL) wfilt = wavelet_prepare (wtype);
		dwt_2d (&mainimg, W_FORWARD, decomp_level, wfilt, *subband);
		wavelet_shrink (&mainimg, decomp_level, wshrinkfunc, tmult);
		dwt_2d (&mainimg, W_INVERSE, decomp_level, wfilt, *subband);
	}

	make_imlib_buffer_from_main_image (cur_slice);
	redraw(0);

	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void wvlt_set_func_menu (GtkWidget *widget, gpointer user_data)
{
int func;

	func = *(int*) user_data;
	switch (func)
	{
		case 0:								/* Forward DWT */
				gtk_widget_hide (wvlt_label4); gtk_widget_hide (wvlt_shrinkmenu);
				gtk_widget_show (wvlt_label1); gtk_widget_show (wvlt_typemenu);
				gtk_widget_hide (wvlt_label5); gtk_widget_hide (wvlt_tmult);
				idwt=W_FORWARD; dp (2, "Changed IDWT flag to forward DWT\n");
				break;
		case 1:								/* Inverse DWT */
				gtk_widget_hide (wvlt_label4); gtk_widget_hide (wvlt_shrinkmenu);
				gtk_widget_show (wvlt_label1); gtk_widget_show (wvlt_typemenu);
				gtk_widget_hide (wvlt_label5); gtk_widget_hide (wvlt_tmult);
				idwt=W_INVERSE; dp (2, "Changed IDWT flag to inverse DWT\n");
				break;
		case 2:								/* Wavelet shrinkage */
				gtk_widget_show (wvlt_label4); gtk_widget_show (wvlt_shrinkmenu);
				gtk_widget_hide (wvlt_label1); gtk_widget_hide (wvlt_typemenu);
				gtk_widget_show (wvlt_label5); gtk_widget_show (wvlt_tmult);
				break;
		case 3:								/* Wavelet shrinkage */
				gtk_widget_show (wvlt_label4); gtk_widget_show (wvlt_shrinkmenu);
				gtk_widget_show (wvlt_label1); gtk_widget_show (wvlt_typemenu);
				gtk_widget_show (wvlt_label5); gtk_widget_show (wvlt_tmult);
				break;
	}

	woperation = func;
}


void wvlt_set_shrink_menu (GtkWidget *widget, gpointer user_data)
{

	wshrinkfunc = *(int*) user_data;
	dp (2, "Changed shrinking threshold function to %d\n", wshrinkfunc);
	
}


void wvlt_set_type_menu (GtkWidget *widget, gpointer user_data)
{

	wfunction = *(int*) user_data;
	dp (2, "Changed wavelet to %d\n", wfunction);
	
}

void on_wvlt_cancel_clicked  (GtkButton *button, gpointer user_data)
{

	gtk_widget_destroy (GTK_WIDGET(user_data));
}


/* "Show" performs an inverse transform on a unit vector and thus generates the wavelet itself */

void on_wvlt_show_clicked  (GtkButton *button, gpointer user_data)
{
int M=2048;				/* lb 2048 = 11 levels */
float xdata[M], ydata[M+M];
int i;
WAVELET_TYPE wtype;

	wtype = set_subband_ext (wfunction);
	if (wtype!=W_NULL) wfilt = wavelet_prepare (wtype);
	for (i=0; i<M; i++) {xdata[i]=i; ydata[i]=0; ydata[M+i]=0; }
	ydata[M+8] = 10;		/* Generates wavelet at 2 levels fewer than maximum - green */
	ydata[0] = 10; 			/* Generates scaling fn at 2 levels fewer than maximum - red */
	dwt_1dn (ydata, M, W_INVERSE, 8, wfilt, (*subband));
	dwt_1dn (&ydata[M], M, W_INVERSE, 8, wfilt, (*subband));
	multi_graphme (xdata, ydata, M, 2);

}


/* "Test" performs a full DWT/IDWT on a fixed 1D test signal */

void on_wvlt_test_clicked  (GtkButton *button, gpointer user_data)
{
WAVELET_TYPE wtype;
float tmult;

	tmult = wvlt_tmult_adj->value;

	wtype = set_subband_ext (wfunction);
	if (wtype!=W_NULL) wfilt = wavelet_prepare (wtype);
	dp (1, "Preparing 1D test with wavelet %d\n", wtype);
	test_wavelet (woperation, tmult, wfunction, wshrinkfunc, wfilt, (*subband));

}




void create_wvlt_window (image_type *img)
{
  GtkWidget *wvlt_window;
  GtkWidget *vbox1, *hbox1, *wvlt_table;
  GtkWidget *wvlt_label2;
  GtkWidget *wvlt_label3;
  GtkWidget *hsep1;
  GtkWidget *wvlt_okbutton;
  GtkWidget *wvlt_showbutton;
  GtkWidget *wvlt_testbutton;
  GtkWidget *wvlt_cancelbutton;
  GtkWidget *wvlt_typemenu_menu;
  GtkWidget *wvlt_funcmenu_menu;
  GtkWidget *wvlt_shrinkmenu_menu;
  GtkWidget *glade_menuitem;
  GtkWidget *wvlt_decomp;
  GtkTooltips *tt_wvltcancel, *tt_wvltdecomp, *tt_wvltfn, *tt_wvltok, *tt_wvltshow, *tt_wvltshrink, *tt_wvlttest, *tt_wvlttmult, *tt_wvlttype;

  float min,max;
  int y;

	imgminmax (img, &min, &max);

	wvlt_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (wvlt_window), "wvlt_window", wvlt_window);
	gtk_window_set_title (GTK_WINDOW (wvlt_window), "Wavelet Functions (Experimental)");

	vbox1 = vbox_in_window (wvlt_window,"vbox1"); 
	wvlt_table = MakeTable (wvlt_window, vbox1, "table1", 2,6);
	y=0;

	wvlt_label3 = MakeLabel (wvlt_window,wvlt_table,"Function","wvlt_label3",0,y);
	wvlt_funcmenu = Make_Optionmenu (wvlt_window, wvlt_table, "wvlt_funcmenu", 1,y);
	wvlt_funcmenu_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (wvlt_funcmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_func_menu), "Forward DWT", (gpointer) &menuopt0);
	glade_menuitem = Option_menuitem (wvlt_funcmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_func_menu), "Inverse DWT", (gpointer) &menuopt1);
	glade_menuitem = Option_menuitem (wvlt_funcmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_func_menu), "Wavelet Shrinking", (gpointer) &menuopt2);
	glade_menuitem = Option_menuitem (wvlt_funcmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_func_menu), "All-in-one Denoising", (gpointer) &menuopt3);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (wvlt_funcmenu), wvlt_funcmenu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (wvlt_funcmenu), 0); 
	gtk_container_set_border_width (GTK_CONTAINER (wvlt_funcmenu), 5);
	y++;
	idwt = 0; woperation=0;
	tt_wvltfn = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_wvltfn, wvlt_funcmenu, "Select operation. Note that\nwavelet shrinking works on a DWT image\nonly. Therefore, wavelet denoising is a\n3-step process: FDWT, shrink, IDWT.", NULL);

	wvlt_label4 = MakeLabel (wvlt_window,wvlt_table,"Threshold","wvlt_label4",0,y);
	wvlt_shrinkmenu = Make_Optionmenu (wvlt_window, wvlt_table, "wvlt_shrinkmenu", 1,y);
	wvlt_shrinkmenu_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_shrink_menu), "Universal/VisuShrink (soft)", (gpointer) &menuopt0);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_shrink_menu), "Universal/VisuShrink (hard)", (gpointer) &menuopt1);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_shrink_menu), "BayesShrink (soft)", (gpointer) &menuopt2);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_shrink_menu), "BayesShrink (hard)", (gpointer) &menuopt3);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_shrink_menu), "SUREShrink (soft)", (gpointer) &menuopt4);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_shrink_menu), "SUREShrink (hard)", (gpointer) &menuopt5);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_shrink_menu), "GCVT (soft)", (gpointer) &menuopt6);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_set_shrink_menu), "GCVT (hard)", (gpointer) &menuopt7);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (wvlt_shrinkmenu), wvlt_shrinkmenu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (wvlt_shrinkmenu), 0); idwt = 0;
	gtk_container_set_border_width (GTK_CONTAINER (wvlt_shrinkmenu), 5);
	gtk_widget_hide (wvlt_label4); gtk_widget_hide (wvlt_shrinkmenu);
	y++;
	tt_wvltshrink = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_wvltshrink, wvlt_shrinkmenu, "Wavelet shrinking operation or threshold operation", NULL);

	wvlt_label1 = MakeLabel (wvlt_window,wvlt_table,"Wavelet type","wvlt_label1",0,y);
	wvlt_typemenu = Make_Optionmenu (wvlt_window, wvlt_table, "wvlt_typemenu", 1,y);
	wvlt_typemenu_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "Haar", (gpointer) &menuopt0);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "Daubechies-4", (gpointer) &menuopt1);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "Daubechies-8", (gpointer) &menuopt2);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "Daubechies-12", (gpointer) &menuopt3);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "Daubechies-20", (gpointer) &menuopt4);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "Coiflet-6", (gpointer) &menuopt5);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "Coiflet12", (gpointer) &menuopt6);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "Symlet-4", (gpointer) &menuopt7);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "Symlet-8", (gpointer) &menuopt8);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "Symlet-12", (gpointer) &menuopt9);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "Haar (standalone)", (gpointer) &menuopt10);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_set_type_menu), "DAUB4 (standalone)", (gpointer) &menuopt11);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (wvlt_typemenu), wvlt_typemenu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (wvlt_typemenu), 0); wfunction=0;
	gtk_container_set_border_width (GTK_CONTAINER (wvlt_typemenu), 5);
	y++;
	tt_wvlttype = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_wvlttype, wvlt_typemenu, "Choose mother wavelet", NULL);

	wvlt_label2 = MakeLabel (wvlt_window,wvlt_table,"Decomposition level","wvlt_label2",0,y);
	wvlt_decomp_adj = GTK_ADJUSTMENT(gtk_adjustment_new (3, 1, 10, 1, 1, 0));
	wvlt_decomp = MakeHScale (wvlt_decomp_adj, wvlt_window, wvlt_table,"wvlt_decomp", 1,y);
	gtk_scale_set_digits (GTK_SCALE (wvlt_decomp), 0);
	y++;
	tt_wvltdecomp = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_wvltdecomp, wvlt_decomp, "Maximum dyadic decomposition levels", NULL);

	wvlt_label5 = MakeLabel (wvlt_window,wvlt_table,"Threshold Multiplier","wvlt_label5",0,y);
	wvlt_tmult_adj = GTK_ADJUSTMENT(gtk_adjustment_new (1, 0, 5, 0.1, 1, 0));
	wvlt_tmult = MakeHScale (wvlt_tmult_adj, wvlt_window, wvlt_table,"wvlt_tmult", 1,y);
	gtk_scale_set_digits (GTK_SCALE (wvlt_tmult), 1);
	gtk_widget_hide (wvlt_label5); gtk_widget_hide (wvlt_tmult);
	y++;
	tt_wvlttmult = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_wvlttmult, wvlt_tmult, "Factor to multiply the shrinking theshold with.\nHigher = stronger shrinking.", NULL);

	
	hsep1 = MakeHsep (wvlt_window, wvlt_table, "hsep1", 0,y);
	hsep1 = MakeHsep (wvlt_window, wvlt_table, "hsep2", 1,y);
	y++;

	wvlt_showbutton = Button_in_table (wvlt_window, wvlt_table, "Show wavelet", "wvlt_showbutton",0,y, NULL);
	wvlt_testbutton = Button_in_table (wvlt_window, wvlt_table, "Test 1D DWT", "wvlt_testbutton",1,y, NULL);
	y++;
	tt_wvltshow = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_wvltshow, wvlt_showbutton, "Reconstruct wavelet and scaling function\nby 1D-IDWT of an impulse signal.", NULL);
	tt_wvlttest = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_wvlttest, wvlt_testbutton, "Function to demonstrate 1D denoising (experimental)", NULL);


	hsep1 = MakeHsep (wvlt_window, vbox1, "hsep3", -1,-1);
	hbox1 = hbox_in_window (wvlt_window, vbox1, "wvlt_hbox");

	wvlt_okbutton = Button_in_table (wvlt_window, hbox1, "OK", "wvlt_okbutton",-1,-1, NULL);
	wvlt_cancelbutton = Button_in_table (wvlt_window, hbox1, "CANCEL", "wvlt_cancelbutton",-1,-1,NULL);
	tt_wvltok = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_wvltok, wvlt_okbutton, "Execute the selected operation", NULL);
	tt_wvltcancel = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_wvltcancel, wvlt_cancelbutton, "Abort operation, close window.", NULL);

	gtk_signal_connect (GTK_OBJECT (wvlt_okbutton), "clicked", GTK_SIGNAL_FUNC (on_wvlt_ok_clicked), wvlt_window);
	gtk_signal_connect (GTK_OBJECT (wvlt_showbutton), "clicked", GTK_SIGNAL_FUNC (on_wvlt_show_clicked), wvlt_window);
	gtk_signal_connect (GTK_OBJECT (wvlt_testbutton), "clicked", GTK_SIGNAL_FUNC (on_wvlt_test_clicked), wvlt_window);
	gtk_signal_connect (GTK_OBJECT (wvlt_cancelbutton), "clicked", GTK_SIGNAL_FUNC (on_wvlt_cancel_clicked), wvlt_window);
	gtk_signal_connect (GTK_OBJECT (wvlt_window), "destroy", GTK_SIGNAL_FUNC (on_wvlt_cancel_clicked), wvlt_window);


	gtk_widget_show (wvlt_window); 

}


void wavelet_main ()
{

	create_wvlt_window (&mainimg);

}



/**************************************************************

	The following is a utility function that can be used
	by other modules to perform 1D wavelet denoising.


**************************************************************/





/**************************************************************

	This is the macro interpreter function that hooks up
	to the macro interpreter. A pointer to this function gets
	passed to the macro registry function.


**************************************************************/

void dwt_macro_driver (image_type *img, WAVELET_TYPE wvlt, int idwt, int level)
{

	subband = &general_subband;			/* We always use the general_subband function in macros */
	wfilt = wavelet_prepare (wvlt);
	
	dwt_2d (img, idwt, level, wfilt, *subband);
	make_minmax (img);
}


/*
	DWT2D function
	Syntax: imgvar = DWT2D (imgexpr, "wavelet", idwt, level);
	Compute the discrete wavelet transform of imgexpr.
	"wavelet" is the wavelet type, one of HAAR, DAUB4, DAUB12, DAUB20...
	idwt decides between forward transform (idwt=0) and inverse transform (idwt=1)
	level is the decomposition level.
*/

variable_type mac_dwt (unsigned char* s, int* i)
{
variable_type v,r;
int j,k;
char wparm[20];
WAVELET_TYPE wvlt;
int idwt, level;


	/* The image stack */

	j = *i; k=0; v = nullvar;
	if (!isalpha(s[j])) macro_error (1,j);

	v = expr (s,&j, MT_IMG);		/* Get image to transform */
	
	/* The wavelet parameter */	

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_STRING);	strncpy (wparm, r.strg, 18); free (r.strg);
	wvlt = W_NULL;
	if (!strncasecmp (wparm, "haar",4)) wvlt = W_HAAR;
	else if (!strncasecmp (wparm, "daub4",5)) wvlt = W_DAUB4;
	else if (!strncasecmp (wparm, "daub8",5)) wvlt = W_DAUB8;
	else if (!strncasecmp (wparm, "daub12",6)) wvlt = W_DAUB12;
	else if (!strncasecmp (wparm, "daub20",6)) wvlt = W_DAUB20;
	else if (!strncasecmp (wparm, "coif6",5)) wvlt = W_COIF6;
	else if (!strncasecmp (wparm, "coif12",6)) wvlt = W_COIF12;
	else macro_error (15,j);	

	/* the integer arguments idwt and level */	

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); idwt = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_REAL); level = r.intvalue;

	while (s[j]!=')') j++;
	*i = j+1;

	/* Go execute. Prepare v to be returned as image */

	dwt_macro_driver (&v.imgptr, wvlt, idwt, level);
	v.type=MT_IMG;

	return v;
}



/**************************************************************

	This is the infrastructure we use to register with the 
	main program. The pointer "menufunction" points to the 
	main entry (the function to be called when the menu is 
	activated). Initmodule is the one function that the 
	main program calls (the name is fixed!) to initialize
	the module and to register the module with the menu.

***************************************************************/

void (*menufunction)();
variable_type (*macrofunction)(unsigned char* s, int* i);


void module_main()
{

	dp (1, "Wavelet module activated\n");
	if (file_ok) wavelet_main ();
}




void initmodule ()
{
	dp (1, "Wavelet module: Initializer called\n");
	
	/* Register this module with the main menu */
	
	menufunction = &module_main;
	dso_menu_register ("Wavelet", menufunction);
	
	/* Register the module with the macro interpreter */
	
	macrofunction = &mac_dwt;
	register_dso_macro ("DWT2D",4,MT_IMG,macrofunction);


}


/********************************************************************************/
