
/*
	callbacks.c - callback functions called from the main menu (second part of main menu) 
*/


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */



#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <sys/types.h>
#include <unistd.h>

#include "application.h"
#include "backgnd.h"
#include "binary.h"
#include "callbacks.h"
#include "defs.h"
#include "dlgbox.h"
#include "fft.h"
#include "filter.h"
#include "filter_adaptive.h"
#include "findclust.h"
#include "fractals.h"
#include "functions.h"
#include "graphdsp.h"
#include "imgbuf.h"
#include "import.h"
#include "interface.h"
#include "macro.h"
#include "magnifier.h"
#include "measure.h"
#include "miscop.h"
#include "morphop.h"
#include "processm.h"
#include "roi.h"
#include "secondwnd.h"
#include "segment.h"
#include "snakes.h"
#include "special1.h"
#include "spiderop.h"
#include "statop.h"
#include "support.h"
#include "threshold.h"
#include "tiffstuff.h"
#include "vertanalysis.h"
#include "viz.h"
#include "watershed.h"



/* for the mouse click & drag events */

int drag_x_beg, drag_y_beg, drag_x_end, drag_y_end;
int r_drag_x_beg, r_drag_y_beg, m_drag_x_beg, m_drag_y_beg;
int motion_lastx, motion_lasty;
char button1down=0;
int button2down=0;
int button3down=0;
int vertexdrag;



/* FILE menu */

void on_open1_activate (GtkMenuItem *menuitem, gpointer user_data)
{
gchar* selectedfile;

	ChooseFile("Select file to load:", window1, 0, &selectedfile);
	if (selectedfile)
	{
		dp (1, "File selected: %s\n", selectedfile);
		file_read (selectedfile);
		g_free (selectedfile);
	}

}


void on_open_multiple1_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	do_open_multiple ();
}


void on_import1_activate (GtkMenuItem *menuitem, gpointer user_data)
{
gchar* selectedfile;

	ChooseFile("Select file to import:", window1, 0, &selectedfile);
	if (selectedfile)
	{
		dp (1, "File selected: %s\n", selectedfile);
		import_dialog (selectedfile);
		g_free (selectedfile);
	}

}


void on_rgbload1_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	rgbload_main ();
}


void on_save1_activate  (GtkMenuItem *menuitem, gpointer user_data)
{

	if (file_ok && master_enable) save_main (&mainimg);		
}

void on_close1_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	close_file();
}


void on_prefs1_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	preferences_main ();
}


void on_quit1_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	app_exit ();
}


/* EDIT menu */

void on_undo_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (master_enable) do_undo ();
}


void on_trimimg_activate  (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable)
	{
		trim_data(&mainimg);
		redraw (1);
	}
}

void on_edit2_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable)
		crop_main ();
}


void on_edit3_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable)
		rotzoom_main ();
}

void on_edit4_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable)
		dataconv_main ();
}


void on_edit5_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (master_enable) edit_clickstack (&mainimg);	
}

void on_values_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) editval_main ();	
}


/* VIEW menu */

void on_magnifier_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (have_magnifier)			/* Toggle between magnifier active/inactive */
	{
		destroy_magnifier ();
		have_magnifier = 0;
	}
	else
	{
		display_magnified (&mainimg, mainimg.xmax/2, mainimg.ymax/2, cur_slice);
		have_magnifier = 1;
	}
}



void on_surfplot_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) surface_draw_main ();
}


void on_lookup_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	statusbar_display ("Implementation error: LUT module has been deprecated");
}

void on_colormap_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) cmap_main();
}



void on_viewmip_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) secondary_mip (&mainimg);
}

void on_viewprofile_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) view_profile();
}

void on_viewlinesumx_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) view_linesum_x(&mainimg, cur_slice);
}

void on_viewlinesumy_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) view_linesum_y(&mainimg, cur_slice);
}

void on_viewzslab_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) view_zslab();
}

void on_viewstatop_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) statop_main();
}

void on_viewbc_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) bright_contrast_main();
}


void on_viewcontours_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) 
		show_contour_lines (&mainimg, cur_slice);

}

/* Measure menu */

void on_stats_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) meas_statistics ();
}

void on_meas_dist_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) meas_distance (&mainimg);
}

void on_bound_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) meas_boundary_pts_gui ();
}

void on_clustmeas_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) clust_meas ();
}

void on_clustcnt_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) clustcnt ();
}

void on_fractdim_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) boxdim_main ();
}

/* PROCESS menu */


void on_invrt_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) 
	{
		invert_image (&mainimg);
		make_imlib_buffer_from_main_image (cur_slice);
		redraw (0);
	}
}


void on_skeletonize_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) skel_main ();
}

void on_morphop_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) morph_main ();
}

void on_fillholes_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) 
	{
		fill_holes_main ();
		make_imlib_buffer_from_main_image (cur_slice);
		redraw (0);
	}
}


void on_bcleanup_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) 
	{
		clean_boundary_features_main ();
		make_imlib_buffer_from_main_image (cur_slice);
		redraw (0);
	}
}



void on_centerfeat_activate (GtkButton  *button, gpointer user_data)
{

	if (file_ok && master_enable) 
	{
		center_feat (&mainimg, cur_slice, 0);
		make_imlib_buffer_from_main_image (cur_slice); 
		redraw (0);
	}
}


void on_rebin_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable)
		rebin_main ();
}

void on_uep_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	/* Deprecated. May be reused elseswhere */
}



void on_process2_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable)
	{
		background_main ();
		redraw (1);
	}
}

void on_threshold1_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) threshold_segment (&mainimg);
}

void on_threshold2_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) dual_threshold (&mainimg);
}

void on_threshold3_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) 
	{
		do_thresholding_universal (&mainimg, -1,0,1);	/* ISODATA */
		make_imlib_buffer_from_main_image (cur_slice); 
		redraw (0);
	}
}

void on_threshold4_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) 
	{
		do_thresholding_universal (&mainimg, -1,1,1);	/* Otsu */
		make_imlib_buffer_from_main_image (cur_slice); 
		redraw (0);
	}
}

void on_threshold5_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) 
	{
		do_thresholding_universal (&mainimg, -1,2,1);	/* Entropy */
		make_imlib_buffer_from_main_image (cur_slice); 
		redraw (0);
	}
}


void on_threshold6_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) 
	{
		do_thresholding_universal (&mainimg, -1,3,1);	/* First Valley */
		make_imlib_buffer_from_main_image (cur_slice); 
		redraw (0);
	}
}


void on_threshold8_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) 
	{
		do_thresholding_universal (&mainimg, -1,4,1);			/* Unused */
		make_imlib_buffer_from_main_image (cur_slice); 
		redraw (0);
	}
}


void on_threshold7_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) 
	{
		create_athresh_wnd (&mainimg);						/* Adaptive */
		make_imlib_buffer_from_main_image (cur_slice); 
		redraw (0);
	}
}


void on_process4_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) clustfind ();
}

void on_runlength_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) runlength_main ();
}

void on_convolve_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) convolve_filter_main ();
}

void on_afilter_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) afilter_filter_main ();
}

void on_fft_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) 
	{
		fft_main (0);
	}
		
}

void on_ifft_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) 
	{
		fft_main (1);
	}
		
}

void on_dht_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) 
	{
		fft_main (2);
	}
		
}

void on_idht_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) 
	{
		fft_main (3);
	}
		
}



void on_fft_rose_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) 
	{
		roseplot_main ();
	}
		
}


void on_fourierfilt_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) fourierfilt_main (&mainimg);
}


void on_waveletfilt_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) wavelet_filter_main ();
}


void on_process6_activate  (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable)
	{
		histeq_main();
	}
}


void on_imgmath_activate  (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable)
	{
		imgmath_main();
	}
}


void on_hough_activate  (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable)
	{
		hough_main();
	}
}


/* SEGMENT menu */

void on_spiderop_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	dp (1, "Spider Op Segmentation \n");
	if (file_ok && master_enable) spider_main (0);
}

void on_snakeop_activate_0 (GtkMenuItem *menuitem, gpointer user_data)
{
	dp (1, "Snake Segmentation - Greedy snake (Haidekker)\n");
	if (file_ok && master_enable) snake_main (0);
}

void on_snakeop_activate_1 (GtkMenuItem *menuitem, gpointer user_data)
{
	dp (1, "Snake Segmentation - Greedy snake (Williams & Shah)\n");
	if (file_ok && master_enable) snake_main (1);
}

void on_snakeop_activate_2 (GtkMenuItem *menuitem, gpointer user_data)
{
	dp (1, "Snake Segmentation - force balance formulation\n");
	if (file_ok && master_enable) snake_main (2);
}


void on_hystthresh_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	dp (1, "Hysteresis Threshold Segmentation \n");
	if (file_ok && master_enable) hysteresis_threshold_main (&mainimg);

}


void on_regiongrow_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	dp (1, "Region Growing Segmentation \n");
	if (file_ok && master_enable) regiongrow_main();
}

void on_growmerge_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	dp (1, "Grow-Merge Segmentation \n");
	if (file_ok && master_enable) grow_merge_main();

}

void on_kmeans_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	dp (1, "k-Means clustering Segmentation \n");
	if (file_ok && master_enable) k_means_main();

}

void on_hillclimb_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	dp (1, "Hill Climbing Segmentation \n");
	if (file_ok && master_enable) hillclimbing_main();

}


void on_watershed_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	dp (1, "Watershed Segmentation \n");
	if (file_ok && master_enable) watershed_main ();

}


/* SPECIAL 1 menu */

// void on_wedge_activate (GtkMenuItem *menuitem, gpointer user_data)
// {
// 	if (file_ok) wedge_main ();
// }
// 
// void on_bonedens_activate (GtkMenuItem *menuitem, gpointer user_data)
// {
// 	if (file_ok) bonedens_main ();
// }

void on_vertebrae_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) spider_main (1);
}


void on_vesselanalysis_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (file_ok && master_enable) spider_main (2);
}


/* MACROS menu */

void on_loadmacro_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	load_macro (NULL);
}

void on_delmacro_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	delete_macros ();
}

void on_immdtmacro_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	immediate_exec ();
}


/* HELP menu */

void on_help1_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	start_help ();
}

void on_about1_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	create_aboutdialog1 ();
}

void on_imginfo1_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	if (master_enable)
	{
		if (file_ok)
			imginfo (&mainimg);
		else
			ShowMessage ("No file opened");
	}
}


void on_license1_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	show_full_license ();
}

void on_relnotes_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	display_release_notes ();
}



void on_pidinfo_activate (GtkMenuItem *menuitem, gpointer user_data)
{
char s[255];
pid_t pid = getpid ();

	if (master_enable)
	{
		sprintf (s, "PID %d",pid);
		statusbar_display (s);
		if (mainimg.parent) set_display_title (s);
		if (secondaryimage_exists) second_statusbar_display (s);
		g_window_set_title (s);
	}
}




/*!
\addtogroup mainlevel

\brief The three buttons

	The three buttons.<p>
	 
	 In earlier versions, the buttons didn't have a special function.
	 I used them to hook up new algorithms or temporary algorithms
	 without having to change the menu structure.<p>
	For the final release, I have assigned the LOAD
	and SAVE functions to two buttons and used the third to toggle
	the thumbnail display. Looks good to me.
	 
*/


void on_button1_clicked (GtkButton  *button, gpointer user_data)
{
gchar* selectedfile;

	if (master_enable)
	{
		ChooseFile("Select file to load:", window1, 0, &selectedfile);
		if (selectedfile)
		{
			dp (1, "File selected: %s\n", selectedfile);
			file_read (selectedfile);
			g_free (selectedfile);
		}
	}
}


void on_button2_clicked (GtkButton *button, gpointer  user_data)
{

	if (file_ok && master_enable) save_main (&mainimg);

}



void on_button3_clicked (GtkButton  *button, gpointer user_data)
{

	/* Toggle the thumbnail image */
	
	if (have_thumbnail)
		destroy_thumbnail ();
	else
		display_thumbnail ();

}


void on_stopbutton_clicked (GtkButton *button, gpointer  user_data)
{

	master_abort = 1;
	if (!macrorun) statusbar_display ("The Stop button is not yet fully implemented");

}


void on_slice_spinbutton_changed (GtkButton *button, gpointer user_data)
{
int z;

	if (file_ok && master_enable)
	{
		z = gtk_spin_button_get_value_as_int ((GtkSpinButton*)user_data);
		new_zslice (z);
	}

}


void on_zoom_changed (GtkButton *button, gpointer user_data)
{
int zoom;

	zoom = (int)gtk_adjustment_get_value ((GtkAdjustment*)user_data);
	dp (2, "New zoom level: %d\n", zoom);


	if (file_ok && master_enable)
	{
		zoomlevel_x = zoomlevel_y = zoom;
		make_imlib_buffer_from_main_image (cur_slice);
		redraw (0);
		if (have_magnifier) display_magnified (&mainimg, mag_lastx, mag_lasty, cur_slice);
	}

}





/********************************************************+

	DISPLAY WINDOW SIGNAL HANDLERS
	
**********************************************************/

/*!

\addtogroup mainlevel

\brief Mouse functions and keys in the image window


Once you have an open image (the main image window), a series of
mouse and key functions are available.

<ul>
<li>	Left-click: 	Display the image value under the crosshairs in the statusbar
<li>	Left-drag: 		Drag a rectangular ROI
<li>	Middle-drag;	Move the ROI
<li>	Right-click:	Remove ROI (or close polygonal ROI)
<li>	Shift-click:	Set a vertex for a polygonal ROI
<li>	Arrow keys:		Move ROI by one pixel
<li>	c and C keys:	Create a circular ROI and shrink/enlarge it
<li>	a key:			"Select All" ROI over the entire image
<li>	A key:			Re-aspect the window
<li>	PgUP key:		Next slice in stack
<li>	PgDN key:		Previous slice in stack
<li>	e key:			Shortcut for <i>Edit->Values</i>
<li>	r and R keys:	Add/remove a snake repulsor
<li>	s and S keys:	Add/remove a snake attractor
<li>	F key:			If pressed above a cluster, return Fourier shape descriptors and others
<li>	f key:			Flood-fill an area. If used above background, fills with Imax. If used above foreground, fills with zero.
</ul>


*/


void signal_close_file ()
{
	if (master_enable)
	{
		close_file ();
		dp (1, "Main window closed\n");
		gtk_widget_destroy (mainimg.parent);
		mainimg.parent = NULL;
	}
}


/* This callback function is called when a pixbuf is referenced to zero. It is used to
	free the memory and set the pointer to zero.
	This needs to be rechecked inn a quiet moment. Do we really need to explicitly free
	the pixelpointer? I think not (and I think this led to crashes in the past), so I
	commented it out (July 2012), but we need to watch for memory leaks. */

void pixbuf_destroy_handler (guchar *pixelptr, void *user_data)
{
GdkPixbuf *pixbuf;

	pixbuf = (GdkPixbuf*)user_data;		/* A pointer to the pixbuf we are dealing with */
	dp (2, "Deactivated destroy handler: pixbuf %08X, data at %08X\n",pixbuf, pixelptr);
/*	if (pixelptr)
	{
		free (pixelptr);
		pixelptr = NULL;
	}*/
}



gboolean on_drawingarea1_expose_event (GtkWidget *widget, GdkEventExpose *event, gpointer user_data)
{
int x,y;		/* To save some typing */
int w,h;

	if(event->count > 0) 			/* Draw only last expose */
	{
		return TRUE;
	}


	x = event->area.x; y = event->area.y;
	w = event->area.width; h = event->area.height;

	/* If anything is out of range, force it back. This can happen if the display
		window is dragged to a size larger than the pixbuf area */

	if (x>mainimg.width)
	{
		x = 0; w = mainimg.width;
	}
	if (y>mainimg.height)
	{
		y = 0; h = mainimg.height;
	}

	if (w>mainimg.width) w=mainimg.width;
	if (h>mainimg.height) h=mainimg.height;
	if (x+w>mainimg.width) w=mainimg.width-x;
	if (y+h>mainimg.height) h=mainimg.height-y;

	dp (3, "Main window expose event. Area is %d, %d, %d, %d\n", x, y, event->area.width, event->area.height);

	gdk_draw_pixbuf (mainimg.drawable->window,
									mainimg.drawable->style->fg_gc[GTK_STATE_NORMAL], 
									mainimg.pixbuf, 
									x,y,							/* src_x and src_y */
									x,y, 							/* dest_x and dest_y */
									w,h,							/* width & height */
									GDK_RGB_DITHER_NORMAL, 0, 0);


  	return FALSE;
}


gboolean on_drawingarea1_configure_event (GtkWidget *widget, GdkEventConfigure *event, gpointer  user_data)
{
int x,y;

	dp (3, "Main window configure event. Size is %d, %d\n", event->width, event->height);

	make_imlib_buffer_from_main_image (cur_slice);

	init_scrollbars ();
	return FALSE;
}


/* This function is called whenever the main window needs to be resized */

void mainwindow_resize_callback (image_type *img, int newx, int newy)
{

	if (img->drawable)
	{
		dp (2, "main window resize callback sets new size %d,%d\n", newx, newy);
		img->width = newx;
		img->height = newy;
		gtk_widget_set_size_request (GTK_WIDGET(img->drawable), newx, newy);
	}
	else
	{
		dp (2, "main window resize callback failed: NULL drawable.\n");
	}

}



/*	Some special key vals
	Event->keyval is an int which contains the ASCII value
	of the key, or a special code for extended keys. Some of them are

		Left arrow		65361 0xff51
		Up arrow		65362 0xff52
		Right arrow		65363 0xff53
		Down arrow		65364 0xff54
		Pg up			65365 0xff55
		Pg down			65366 0xff56
*/

gboolean on_drawingarea1_key_press_event (GtkWidget *widget, GdkEventKey *event, gpointer user_data)
{
int xp, yp,i,j;
float r,rmin;
char buf[257];
GdkModifierType state;

	dp (2, "Key press event: Key value = %d\n", event->keyval);

	if (!master_enable) return FALSE;		/* No keypress handling allowed if a task is running */

	gdk_window_get_pointer (widget->window, &xp, &yp, &state);

	if ((event->keyval >= 0x20) && (event->keyval <= 0xFF))
	{
		switch (event->keyval)
		{
			case 'a':				/* Decrease size to match aspect not allowed by gdk, so use this as... */
				if (!current_roi.roi_mode)		/* select->all (provided there is no ROI) */
				{
					current_roi.xi[0]=0;	current_roi.xi[1]=mainimg.xmax-1;	current_roi.xi[2]=-1;
					current_roi.yi[0]=0;	current_roi.yi[1]=mainimg.ymax-1;	current_roi.yi[2]=-1;
				
					current_roi.bbox_x1 = 0;	/* Bounding box is the same */
					current_roi.bbox_x2 = mainimg.xmax-1;
					current_roi.bbox_y1 = 0;
					current_roi.bbox_y2 = mainimg.ymax-1;
				
					current_roi.roi_mode = ROI_RECT;	/* rectangular ROI - enables marching ants */
				}
				else
					statusbar_display ("Please delete current ROI first");
				break;
			
			case 'A':				/* Increase size to match aspect -- DEPRECATED*/
				re_aspect (1);
				break;
			
			case 'c':				/* Set circular ROI or decrease its size */
			
				if (current_roi.roi_mode==3)
				{
					if (current_roi.radius>1) 
						roi_set_circular_roi (current_roi.xc,current_roi.yc, current_roi.radius-1);
				}
				else
				{
					roi_imgxy (&xp, &yp);
					roi_set_circular_roi (xp,yp,10);
				}
				break;
		
			case 'C':				/* Set circular ROI or increase its size */
			
				if (current_roi.roi_mode==3)
				{
					roi_set_circular_roi (current_roi.xc,current_roi.yc, current_roi.radius+1);
				}
				else
				{
					roi_imgxy (&xp, &yp);
					roi_set_circular_roi (xp,yp,10);
				}
				break;
				
			case 'e':
				if (current_roi.roi_mode) editval_main ();		/* A shortcut for edit->Values */
				break;

			case 'r':
				if ((snake_active) && (nrepulsors<9))			/* Add a snake repulsor */
				{
					roi_imgxy (&xp, &yp);
					snake_repulsors[nrepulsors].x=xp;
					snake_repulsors[nrepulsors].y=yp;
					sprintf (buf, "Added repulsor at (%d,%d)",xp,yp);
					dp (1, "%s\n",buf);
					statusbar_display (buf);
					nrepulsors++;
					snake_interaction_callback (buf);
				}
				break;	
			case 'R':
				if ((snake_active) && (nrepulsors>0))			/* Remove a snake repulsor */
				{
					roi_imgxy (&xp, &yp);
					/* find closest and remove only if nearer that 20 pixels */
					rmin=999999; j=nrepulsors-1;
					for (i=0; i<nrepulsors; i++)
					{
						r = SQR(xp-snake_repulsors[i].x)+SQR(yp-snake_repulsors[i].y);
						if (r<rmin)
						{
							rmin=r; j=i;
						}
					}
					r = SQR(xp-snake_repulsors[j].x)+SQR(yp-snake_repulsors[j].y);	/* found one? */
					if (r<100)		/* closer than 10 pixels */
					{
						sprintf (buf,"Removed repulsor at (%d,%d)",xp,yp);
						dp (1, "%s\n",buf);
						statusbar_display (buf);
						if ((nrepulsors==1) || (j==nrepulsors-1))
							nrepulsors--;		/* Single repulsor or last in list? Just take it away */
						else
						{
							for (i=j+1; i<nrepulsors; i++) snake_repulsors[j]=snake_repulsors[i];
							nrepulsors--;
						}
					}

					snake_interaction_callback (buf);
				}
				break;
			case 's':
				if ((snake_active) && (nattractors<9))			/* Add a snake attractor or (S)pring */
				{
					roi_imgxy (&xp, &yp);
					snake_attractors[nattractors].x=xp;
					snake_attractors[nattractors].y=yp;
					sprintf (buf, "Added attractor (spring) at (%d,%d)",xp,yp);
					dp (1, "%s\n",buf);
					statusbar_display (buf);
					nattractors++;
					snake_interaction_callback (buf);
				}
				break;	
			case 'S':
				if ((snake_active) && (nattractors>0))			/* Remove a snake attractor */
				{
					roi_imgxy (&xp, &yp);
					/* find closest and remove only if nearer that 20 pixels */
					rmin=999999; j=nattractors-1;
					for (i=0; i<nattractors; i++)
					{
						r = SQR(xp-snake_attractors[i].x)+SQR(yp-snake_attractors[i].y);
						if (r<rmin)
						{
							rmin=r; j=i;
						}
					}
					r = SQR(xp-snake_attractors[j].x)+SQR(yp-snake_attractors[j].y);	/* found one? */
					if (r<100)		/* closer than 10 pixels */
					{
						sprintf (buf,"Removed attractor at (%d,%d)",xp,yp);
						dp (1, "%s\n",buf);
						statusbar_display (buf);
						if ((nattractors==1) || (j==nattractors-1))
							nattractors--;		/* Single attractor or last in list? Just take it away */
						else
						{
							for (i=j+1; i<nattractors; i++) snake_attractors[j]=snake_attractors[i];
							nattractors--;
						}
					}

					snake_interaction_callback (buf);
				}
				break;

			case 'F':				/* Cluster Fourier descriptors and other shape descriptors */
					roi_imgxy (&xp, &yp);
					describe_cluster (xp, yp);
					break;
			case 'f':				/* Manual fill-holes. If this is a zero-value pixel, flood-fill with imax */
			{
				int level=0;
				float min, max;

					roi_imgxy (&xp, &yp);
					undo_nexus (&mainimg);
					imgminmax (&mainimg, &min, &max);
					rg_floodfill (&mainimg, max, xp,yp,cur_slice, &level);				/* Floodfill the background */
					make_imlib_buffer_from_main_image (cur_slice);
					redraw (1);
					break;
			}
		
		}
	}
	else if (event->keyval==65365) /* Pg up */
	{
		new_zslice (cur_slice+1);
		gtk_spin_button_set_value (GTK_SPIN_BUTTON(slice_spinbutton), cur_slice);
	}
	else if (event->keyval==65366) /* pg dn */
	{
		new_zslice (cur_slice-1);
		gtk_spin_button_set_value (GTK_SPIN_BUTTON(slice_spinbutton), cur_slice);
	}
	/* Wouldn't we want to extend roi move through arrow keys for all ROI types?? */
	else if ((event->keyval == 65362 /* up arrow key */) && (current_roi.roi_mode==ROI_CIRCLE))
	{
		int x,y,c;
		
		x = current_roi.xc;
		y = current_roi.yc-1;
		c = current_roi.radius;
		kill_roi ();
		roi_set_circular_roi (x,y,c);
	}
	else if ((event->keyval == 65364 /* down arrow key */) && (current_roi.roi_mode==ROI_CIRCLE))
	{
		int x,y,c;
		
		x = current_roi.xc;
		y = current_roi.yc+1;
		c = current_roi.radius;
		kill_roi ();
		roi_set_circular_roi (x,y,c);
	}
	else if ((event->keyval == 65363 /* right arrow key */) && (current_roi.roi_mode==ROI_CIRCLE))
	{
		int x,y,c;
		
		x = current_roi.xc+1;
		y = current_roi.yc;
		c = current_roi.radius;
		kill_roi ();
		roi_set_circular_roi (x,y,c);
	}
	else if ((event->keyval == 65361 /* left arrow key */) && (current_roi.roi_mode==ROI_CIRCLE))
	{
		int x,y,c;
		
		x = current_roi.xc-1;
		y = current_roi.yc;
		c = current_roi.radius;
		kill_roi ();
		roi_set_circular_roi (x,y,c);
	}

	
	return TRUE;
}


/* Gets called when the user presses a mouse button. Note that the event type contains additional information, 
	for example (event->type==GDK_2BUTTON_PRESS) is true for a double-click, and
    (event->type==GDK_2BUTTON_PRESS) is true for a triple-click.
*/

gboolean on_drawingarea1_clicked_event (GtkWidget *widget, GdkEventKey *event, gpointer user_data)
{
int x,y;
GdkModifierType state;

	gdk_window_get_pointer (widget->window, &x, &y, &state);
	dp (1, "*click* at (%d,%d), state is %04x\n",x,y,state);
	if (!master_enable) return FALSE;		/* No mouse click handling allowed if a task is running */
	
	motion_lastx=x; motion_lasty=y;
	
	if (state & GDK_BUTTON1_MASK)		/* left button */
	{
		drag_x_beg = x; drag_y_beg = y;
		button1down=1;
	}
	if (state & GDK_BUTTON3_MASK)		/* right button */
	{
		r_drag_x_beg = x; r_drag_y_beg = y;
		button2down=1;
	}
	if (state & GDK_BUTTON2_MASK)		/* middle button */
	{
		m_drag_x_beg = x; m_drag_y_beg = y;
		button3down=1;
		dragroi_ini=0;
		
		/* If we are near a vertex indicate that we'll drag a vertex and not the ROI */
		
		vertexdrag=0;
		if ((vertexdrag=roi_is_near_vertex(x,y))) 	/* Yes, this is an assignment!!! */
		{
			gdk_window_set_cursor (mainimg.drawable->window, grab_cursor);
			dp (2, "Grabbed vertex %d at mouse coordinates %d,%d\n",vertexdrag,x,y);
		}
		
	}
	
  	return FALSE;
}


gboolean on_drawingarea1_click_released_event (GtkWidget *widget, GdkEventKey *event, gpointer user_data)
{
int x,y;
GdkModifierType state;

	if (!master_enable) return FALSE;		/* No mouse click handling allowed if a task is running */
	gdk_window_get_pointer (widget->window, &x, &y, &state);

	/* Button_released cannot differentiate between buttons */

	drag_x_end = x; drag_y_end = y;

   	
	if ((state & GDK_SHIFT_MASK) && (button1down))	/* shift-click adds polygonal */
	{
		roi_add_vertex (x,y);
	}
	else if ((drag_x_beg==x) && (drag_y_beg==y) && (button1down))
	{
		if (file_ok) clickhandler (x,y);	/* Simple click is pixelpeek */
	}
	else if (button1down)	/* Otherwise, end of dragging a ROI */
	{
		roi_erase_rectangle ();
		roi_set_rectangular (drag_x_beg, drag_y_beg, x, y);
	}
	
	if (button2down) 	/* right button released */
	{
		if ((current_roi.roi_mode==ROI_RECT) || (current_roi.roi_mode==ROI_CIRCLE) || roi_is_closed_polygon())
			kill_roi();	/* Right button stops ROI */
		else
			roi_close_polygon(1);
			
		/* In any case, disable the clickstack display */
		display_clickstack=0;
	}

	if (button3down) 	/* Middle button released */
	{
		if (vertexdrag)
		{
			/* Reset cursor */
		
			gdk_window_set_cursor (mainimg.drawable->window, crosshair_cursor);
			x=drag_x_end - m_drag_x_beg;
			y=drag_y_end - m_drag_y_beg;
			roi_offset_vertex (vertexdrag /* the vertex number*/, x, y);
			dp (2, "Released vertex %d at mouse coordinates %d,%d\n",vertexdrag,x,y);
			redraw (0);
		}
		else
		{
		
			roi_erase_roi ();	/* erase the old auxiliary "dragging" roi */
			x=drag_x_end - m_drag_x_beg;
			y=drag_y_end - m_drag_y_beg;
			roi_imgxy (&x,&y);
			roi_offset_roi (x,y);
			redraw (0);
		}
	}

	/* release all button flags */
	button1down=button2down=button3down=0;

	return FALSE;
}	


gint on_drawingarea1_motion_notify_event  (GtkWidget *widget, GdkEventMotion *event, gpointer user_data)
{
int x,y,x1,y1;
GdkModifierType state;	
char s[128];

	/* Hints are a summary of multiple mouse events combined in one event */

	if (!master_enable) return FALSE;		/* No mouse click handling allowed if a task is running */
	if (event->is_hint)
	{
		dp (3, "Motion notify: Hint\n");
		gdk_window_get_pointer (widget->window, &x, &y, &state);
	}
	else	/* individual mouse event */
	{
		dp (3, "Motion notify: Single event\n");
		x = event->x;
		y = event->y;
		state = event->state;
	}
	
	if (state & GDK_BUTTON1_MASK)	/* Dragging w/ mouse btn 1 down */
	{
		dp (3, "Drag: (%d %d) to (%d %d)\n",drag_x_beg, drag_y_beg, x,y);
		roi_drag_rectangle (drag_x_beg, drag_y_beg, x,y);
	}
	
	if (state & GDK_BUTTON2_MASK)	/* Dragging w/ middle button down */
	{
		if (vertexdrag)
		{
			dp (3, "Dragging vertex: (%d %d) to (%d %d)\n",m_drag_x_beg, m_drag_y_beg, x,y);
			x1=x-m_drag_x_beg; y1=y-m_drag_y_beg; roi_imgxy (&x1,&y1);
			roi_drag_vertex (vertexdrag,x1,y1);
			sprintf (s, "Vertex offset %d, %d",x1,y1);
			statusbar_display (s);
			
		}
		else
		{
			dp (3, "Drag middle button: (%d %d) to (%d %d)\n",m_drag_x_beg, m_drag_y_beg, x,y);
			x1=x-m_drag_x_beg; y1=y-m_drag_y_beg; roi_imgxy (&x1,&y1);
			roi_drag_roi (x1,y1);
			sprintf (s, "ROI offset %d, %d",x1,y1);
			statusbar_display (s);
		}
	}


	motion_lastx=x; motion_lasty=y;

	return TRUE;
}


gboolean on_drawingarea1_enter_notify_event (GtkWidget *widget, GdkEventCrossing *event, gpointer user_data)
{
	dp (3, "Drawingarea enter notify event\n");
	return FALSE;
}


gboolean on_drawingarea1_leave_notify_event (GtkWidget *widget, GdkEventCrossing *event, gpointer user_data)
{

	dp (3, "Drawingarea leave notify event\n");
	return FALSE;
}


/* Gets called when the window receives events from the mouse scroll wheel */

gboolean on_drawingarea1_scroll_event (GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{

	if ((long)(event->axes) & 0x100000000)
		dp (3, "Drawingarea scroll event: Scroll down\n");
	else
		dp (3, "Drawingarea scroll event: Scroll up\n");

	return TRUE;		/* Don't pass the event to the parent window */
}


/* As of 0.9.0, this checkbutton no longer exists. Nor do the scrollbars in their original form */

// void on_checkbutton1_toggled (GtkToggleButton *togglebutton, gpointer user_data)
// {
// 	autosize = !autosize;
// 	init_scrollbars ();
// 	redraw (0);
// 
// }


// void on_hscrollbar1_motion (GtkWidget *widget, GdkEventKey *event, gpointer user_data)
// {
// GtkAdjustment* adj;
// int value, scr1, scr2;
// 
// 	adj = (GtkAdjustment*) widget;
// 	value = adj->value;
// 	scr1 = adj->upper - adj->lower - adj->page_size;
// 	scr2 = cur_w-drawingarea1->allocation.width;
// 	
// 	xofs = (int) ( (double)(scr2)*value/(scr1));
// 	if (!autosize) redraw (0);
// }
// 
// void on_vscrollbar1_motion (GtkWidget *widget, GdkEventKey *event, gpointer user_data)
// {
// GtkAdjustment* adj;
// int value, scr1, scr2;
// 
// 	adj = (GtkAdjustment*) widget;
// 	value = adj->value;
// 	scr1 = adj->upper - adj->lower - adj->page_size;
// 	scr2 = cur_h-drawingarea1->allocation.height;
// 	
// 	yofs = (int) ( (double)(scr2)*value/(scr1));
// 	if (!autosize) redraw (0);
// }



/**********************************************************************************************/




