
/*
*	Graphdsp: Functions to display simple graphs y = f(x)
*	e.g. profile plots, histograms
*/


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */



#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "graphdsp.h"
#include "application.h"
#include "dlgbox.h"
#include "functions.h"
#include "functions2.h"


double rxmin, rxmax, rymin, rymax;	/* Data value range */
int fxmin, fxmax, fymin, fymax; /* useable area within graph window, pixels */

float* cxdata = NULL; float* cydata = NULL;
int Datapoints, Datasets;
static int g_statusbar_was_used=0;

GtkWidget *graphwindow;
GtkWidget *graphdsp_drawingarea;
GtkWidget* gstatusbar1;

/* for the mouse click & drag events */

int gdrag_x_beg, gdrag_y_beg, gdrag_x_end, gdrag_y_end;
int gr_drag_x_beg, gr_drag_y_beg, gm_drag_x_beg, gm_drag_y_beg;
int gmotion_lastx, gmotion_lasty;
char gbutton1down=0;
char gbutton2down=0;
char gbutton3down=0;

/* Prototype it because it's used before its declaration */

void g_statusbar_display (const char* text);


/* Some stock colors in rgb format:

	 0: black
	 1: red
	 2: green
	 3: blue
	 4: magenta
	 5: yellow
	 6: cyan
	 7: dark magenta
	 8: grey
	 9: dark red
	10: dark green
	11: dark blue
	12: dark magenta
	13: dark yellow
	14: dark cyan
	15: white
	
*/

rgbtriplet colortable[16] = { {0,0,0}, {0xff,0,0}, {0,0xff,0}, {0,0,0xff}, 
							 {0xff,0,0xff}, {0xff,0xff,0}, {0,0xff,0xff}, 
							 {0x90,0x00,0x90}, {0x7f,0x7f,0x7f}, {0x7f,0,0},
							 {0,0x7f,0}, {0,0,0x7f}, 
							 {0x7f,0,0x7f}, {0x7f,0x7f,0}, {0,0x7f,0x7f}, 
							 {0xff,0xff,0xff} };

/* Light-weight colors */

rgbtriplet ltcolortable[16] = { {0x7f,0x7f,0x7f}, {0xff,0x7f,0x7f}, {0x7f,0xff,0x7f}, {0,0,0xff}, 
							 {0xff,0x7f,0xff}, {0xff,0xff,0x7f}, {0x7f,0xff,0xff}, 
							 {0x90,0x40,0x90}, {0xbf,0xbf,0xbf}, {0x7f,0x3f,0x3f},
							 {0x3f,0x7f,0x3f}, {0x3f,0x3f,0x7f}, 
							 {0x7f,0x3f,0x7f}, {0x7f,0x7f,0x3f}, {0x3f,0x7f,0x7f}, 
							 {0xff,0xff,0xff} };





/************************************************************************/



void determine_scale (float* xdata, float* ydata, int datapoints, int datasets)
{
int i;

	if (datapoints<2) return;
	
	rxmin=rxmax=xdata[0]; rymin=rymax=ydata[0];

	for (i=1; i<datapoints; i++)
	{
		if (rxmin>xdata[i]) rxmin = xdata[i];
		if (rxmax<xdata[i]) rxmax = xdata[i];
	}
	for (i=1; i<datapoints*datasets; i++)
	{
		if (rymin>ydata[i]) rymin = ydata[i];
		if (rymax<ydata[i]) rymax = ydata[i];
	}
	
	if (rxmin==rxmax) { rxmin=0; rxmax=20; }
	if (rymin==rymax) { rymin=0; rymax=20; }
}


void ClearDraw(GtkWidget* draw)
{
int ix,iy;
GdkGC* gc;

	ix = draw->allocation.width;
	iy = draw->allocation.height;
	
	/* Initialize the fx/fy values */
	
	fxmin = 90;	/* Leave some space at the left side for y axis numbering */
	fxmax = ix-20;
	fymin = 20;
	fymax = iy-40;
	
	/* Assign default values to rx/ry */
	
	rxmin=0; rymin=0; rxmax=20; rymax=20;

	gc = gdk_gc_new (draw->window);	/* allocate mem & assign null values */
	gdk_gc_copy (gc, draw->style->white_gc); /* Copy some values */

	gdk_draw_rectangle (draw->window, gc, TRUE, 0,0,ix,iy);
	gdk_gc_destroy (gc);	/* free the memory associated with the gc */

}




int UserX (double x)
{
int ix;
	
	ix = (int) ((x-rxmin)*((double)fxmax-(double)fxmin)/(rxmax-rxmin) + fxmin + 0.5);
	if (ix>fxmax) ix=fxmax;
	else if (ix<fxmin) ix = fxmin;
	return ix;
}

int UserY (double y)
{
int iy;

	iy = (int) (fymax - (y-rymin)*(fymax-fymin)/(rymax-rymin) + 0.5);
	if (iy>fymax) iy=fymax;
	else if (iy<fymin) iy = fymin;
	return iy;
}



double RUserX (int ix)
{
double x;

	x = (double)ix;
	x = (x-fxmin)*(rxmax-rxmin)/(fxmax-fxmin) + rxmin;
	return x;
}

double RUserY (int iy)
{
double y;

	y = (double)iy;
	y = (y-fymin)*(rymax-rymin)/(fymax-fymin) + rymin;
	return y;
}


int sign(double x)
{
	if (x==0) return 0;
	else if (x<0) return -1;
	else return 1;
}


double decade(double x)
{
double z;

	if (x==0)
		return 1;
	else
	{
		z = log10(x);
		z = pow (10, floor(z));
		return z;
	}
}

void minimax(double* pmin, double* pmax)
{
double a,r,d,min,max;
int mantissa1;
	
	min = *pmin; max = *pmax;
	d = decade (max-min); mantissa1 = (int)(floor(0.9999+max/d));
	a = 1.0;
	if (mantissa1==2) a=5;
	else if (mantissa1==3) a=2;

	/* in the case of negative numbers, floor rounds into the wrong direction */

	max = sign(max)*floor(1+fabs(max)*a/d)*d/a; 
	r = sign(min)*floor(fabs(min)/d)*d;
	while (r>min) r-=d;
	min = r;
	*pmin=min; *pmax = max;
}

double calcstep(double min, double max, int numticks)
{
char goodval[3]= {1,2,5};
int i, stepcnt;
double r,d;	

	d = decade (max)/100.0;
	i=0;
	do
	{
		r = min; stepcnt=0;
		do { r += d*goodval[i]; stepcnt++; } while (r<max);
		if (stepcnt>numticks)
		{
			i++; 
			if (i>2) { i=0; d*=10; }
		}
	}
	while (stepcnt>numticks);
	return d*goodval[i];

}


void axislabel_format (char* s, float f)
{
int n;

	n=0; 
	if (f<0) n=1; 

	if (f==0.0)
		sprintf (s, "0.0");
	else if ((fabs(f) < 0.01) || (fabs(f) > 10000.0))
		sprintf (s, "%6.2E", f);
	else if (fabs(f) <1.0)
		sprintf (s, "%6.3f", f);
	else if (fabs(f) <10.0)
		sprintf (s, "%6.2f", f);
	else if (fabs(f) <1000.0)
		sprintf (s, "%6.1f", f);
	else
		sprintf (s, "%6.0f", f);
}


void Frame(GtkWidget* draw, float* xdata, float* ydata, int datapoints, int datasets)
{
float x,y;
float lstepx, lstepy;
int ix,iy;
char s[255];
int ox, oy, fontwidth, fontheight;
int numticks;
GdkGC* gc;
PangoLayout *playout;


	ClearDraw (draw);		/* Clear and initialize */
	determine_scale (xdata, ydata, datapoints, datasets);

	/* Here it gets ugly. Every time I upgrade to a newer OS version, *something*
		has changed. gdk_font_load has been deprecated. gdk_draw_string has been deprecated,
		use gdk_draw_layout() instead. Not to be outdone, the link to gdk_draw_layout reveals that
		gdk_draw_layout is deprecated, use some sort of PangoLayout......
		This is the umpteenth time I had to do some modification to this section of code.
		Would it be SO difficult to just provide a function to place some text in some drawable??
		
		PEOPLE, GET REAL! How do you expect us to keep up with development if every time
		you advance gtk by one minor version, we have to do major rewrites of the UI code?!
		I want a functional program. If it does not have the newest shiny glitzy UI elements,
		you know, SO WHAT?

		Guess what. Many programmers will just find a kludge as workaround, and overall software
		quality suffers. Consider that when you discuss progress, esteemed GTK developers. Thanks.
	*/
	
	gc = gdk_gc_new (draw->window);			/* Extract the graphics context from the current window */

	/* Note that we moved font extraction into the loops and now use (the deprecated) gdk_draw_layout
	   instead of gdk_draw_string. Do we need to free or unref the Pango layout? Probably. */

	/* handle the x and y axis ticks. Use a dummy Pango to get font size */

	sprintf (s, "X");			/* a single dummy character */
	playout = gtk_widget_create_pango_layout (draw, s);
	pango_layout_get_pixel_size (playout, &fontwidth, &fontheight);
	g_object_unref(playout);
	numticks = draw->allocation.width/(6*fontwidth);
	minimax (&rxmin, &rxmax);
	lstepx = calcstep (rxmin, rxmax, numticks);		/* Tick increments along x axis */
	numticks = draw->allocation.height/(2*fontheight);
	minimax (&rymin, &rymax);
	lstepy = calcstep (rymin, rymax, numticks);		/* Tick increments along y axis */


	if ((rxmin<0) && (rxmax>0))		// draw y-axis at x=0
	{
		gdk_draw_line (draw->window, gc, UserX(0),UserY(rymin), UserX(0),UserY(rymax)); 
		gdk_draw_line (draw->window, gc, UserX(0)-1,UserY(rymin), UserX(0)-1,UserY(rymax)); 
	}
	else							// draw y-axis at xmin
	{
		gdk_draw_line (draw->window, gc, UserX(rxmin),UserY(rymin), UserX(rxmin),UserY(rymax)); 
		gdk_draw_line (draw->window, gc, UserX(rxmin)-1,UserY(rymin)-1, UserX(rxmin)-1,UserY(rymax)); 
	}

	if ((rymin<0) && (rymax>0))		// draw x-axis at y=0
	{
		gdk_draw_line (draw->window, gc, UserX(rxmin),UserY(0), UserX(rxmax),UserY(0)); 
		gdk_draw_line (draw->window, gc, UserX(rxmin),UserY(0)+1, UserX(rxmax),UserY(0)+1); 
	}
	else							// draw x-axis at ymin
	{
		gdk_draw_line (draw->window, gc, UserX(rxmin),UserY(rymin), UserX(rxmax),UserY(rymin)); 
		gdk_draw_line (draw->window, gc, UserX(rxmin),UserY(rymin)+1, UserX(rxmax),UserY(rymin)+1); 
	}


	for (x=rxmin; x<=rxmax; x+=lstepx)		/* Ticks and numbers for x axis */
	{
		ix = UserX(x); iy = UserY(0);
		gdk_draw_line (draw->window, gc, ix,iy, ix,iy+5);
		axislabel_format (s,x);
		playout = gtk_widget_create_pango_layout (draw, s);
		pango_layout_get_pixel_size (playout, &ox, &oy);
		gdk_draw_layout (draw->window, gc, ix-ox/2, iy+8, playout);
		g_object_unref(playout);
	}

	for (y=rymin; y<=rymax; y+=lstepy)		/* Ticks and numbers for y axis */
	{
		ix = UserX(0); iy = UserY(y);
		gdk_draw_line (draw->window, gc, ix,iy, ix-5,iy);
		axislabel_format (s,y);
		playout = gtk_widget_create_pango_layout (draw, s);
		pango_layout_get_pixel_size (playout, &ox, &oy);
		gdk_draw_layout (draw->window, gc, ix-ox-15, iy-oy/2, playout);
		g_object_unref(playout);
	}

	gdk_gc_destroy (gc);	/* free the memory associated with the gc */

}


void DrawData(GtkWidget* draw, float* xdata, float* ydata, int datapoints, int datasets)
{
int i,j,x,y;
GdkGC* gc;
GdkColor color;

	gc = gdk_gc_new (draw->window);	/* allocate mem & assign null values */
	gdk_gc_copy (gc, draw->style->black_gc); /* Copy some values */
	
	for (j=0; j<datasets; j++)
	{
		if (datasets>1)
		{
			color.red = colortable[(j+1) & 0x7].red << 8;
			color.blue = colortable[(j+1) & 0x7].blue << 8;
			color.green = colortable[(j+1) & 0x7].green << 8;
			gdk_color_alloc (gdk_colormap_get_system(), &color);
			gdk_gc_set_foreground (gc, &color);
		}
		
		if (datapoints<15)
		{
			for (i=0; i<datapoints; i++)		// draw the individual points as small circles
			{
				dp (2, "%f   %f\n", xdata[i],ydata[i]); 
				x = UserX(xdata[i]); y = UserY(ydata[i+j*datapoints]); 
				gdk_draw_arc (draw->window, gc, TRUE, x-3,y-3, 6,6, 0, (360*64) );
			}
		}

		x = UserX(xdata[0]); y = UserY(ydata[j*datapoints]); 
		for (i=1; i<datapoints; i++)
		{
			gdk_draw_line (draw->window, gc, x,y,UserX(xdata[i]),UserY(ydata[i+j*datapoints]));
			x = UserX(xdata[i]); y = UserY(ydata[i+j*datapoints]); 
		}
	}
	
	gdk_gc_destroy (gc);	/* free the memory associated with the gc */

}



void drawall (GtkWidget* draw)
{

	Frame (draw, cxdata, cydata, Datapoints, Datasets);
	DrawData (draw, cxdata, cydata, Datapoints, Datasets);
}



/* Return the index so that xdata[i] <= x < xdata[i+1] */

int xindex (double x, float* xdata, int datapoints)
{
int i;

	i=0; while ((i<datapoints) && (xdata[i]<=x)) i++;
	i--;

	return i;

}


/* Return interpolated data at x-position x */

double getdata (double x, int setno, float* xdata, float* ydata, int datapoints, int datasets)
{
int i,idx;
double x1,x2,y1,y2;

	/* Index forward until x exceeds xdata[] */
	
	if (setno>=datasets) setno=datasets-1;
	idx = setno*datapoints;
	i = xindex (x,xdata,datapoints);
	if (i>=datapoints-1) return ydata[idx+datapoints-1];
	
	/* The following rather explicit for easier understanding and debugging */

	y1=ydata[idx+i]; y2=ydata[idx+i+1];
	x1=xdata[i]; x2=xdata[i+1];
	return y1 + (y2-y1)*(x-x1)/(x2-x1);
	
}



/* Clicked on the drawing area -> display function value */

void g_clickhandler (int fx, int fy)
{
char s[255], t[255];
double x,y;
int j;
GdkGC *gc;
GdkColor ltgray = {0x9600,0x9600,0x9600};
GdkColor color;

	drawall (graphdsp_drawingarea);  /* Erase previous lines */

	x = RUserX(fx);

	/* Display a thin gray dashed line at the x value */

	gc = gdk_gc_new (graphdsp_drawingarea->window);	/* allocate mem & assign null values */
	gdk_gc_copy (gc, graphdsp_drawingarea->style->black_gc); /* Copy some values */
	gdk_gc_set_dashes (gc, 0, "\3\3", 2);
	gdk_gc_set_line_attributes (gc,1,GDK_LINE_ON_OFF_DASH,GDK_CAP_NOT_LAST,GDK_JOIN_MITER);

	if ((x>=rxmin) && (x<=rxmax))
	{
		gdk_draw_line (graphdsp_drawingarea->window, gc, UserX(x),UserY(rymin),UserX(x),UserY(rymax));


		/* find the x position in the datasets and display values .
		   For each value, draw a line from the x intercept to the y axis */

		sprintf (s, "x=%6.2f ", x);

		for (j=0; j<Datasets; j++)
		{
			y = getdata (x,j,cxdata,cydata,Datapoints,Datasets);
			sprintf (t, " y%d=%6.2f ",j+1,y);
			strcat (s,t);

			color.red=ltcolortable[j].red<<8; color.green=ltcolortable[j].green<<8; color.blue=ltcolortable[j].blue<<8; 
			/*dp (2, "Clickhandler: Setting color %02X%02X%02X\n",ltcolortable[j].red,ltcolortable[j].green,ltcolortable[j].blue);
			gdk_gc_set_foreground (gc, &color);
			gdk_gc_set_background (gc, &color);*/

			gdk_draw_line (graphdsp_drawingarea->window, gc, UserX(x),UserY(y),UserX(rxmin),UserY(y));


		}

		
		g_statusbar_display (s);
	}
	gdk_gc_destroy (gc);


}


/* Shift-Clicked on the drawing area -> display function value of the closest local maximum */

void g_sftclickhandler (int fx, int fy)
{
char s[255], t[255];
double x,y;
int j;
GdkGC *gc;
GdkColor ltgray = {0x9600,0x9600,0x9600};
GdkColor color;

	drawall (graphdsp_drawingarea);  /* Erase previous lines */

	x = RUserX(fx); j = (int)(floor(x));
	x = find_nearby_peak (cydata, Datapoints, j);

	/* Display a thin gray dashed line at the x value */

	gc = gdk_gc_new (graphdsp_drawingarea->window);	/* allocate mem & assign null values */
	gdk_gc_copy (gc, graphdsp_drawingarea->style->black_gc); /* Copy some values */
	gdk_gc_set_dashes (gc, 0, "\3\3", 2);
	gdk_gc_set_line_attributes (gc,1,GDK_LINE_ON_OFF_DASH,GDK_CAP_NOT_LAST,GDK_JOIN_MITER);

	if ((x>=rxmin) && (x<=rxmax))
	{
		gdk_draw_line (graphdsp_drawingarea->window, gc, UserX(x),UserY(rymin),UserX(x),UserY(rymax));


		/* find the x position in the datasets and display values .
		   For each value, draw a line from the x intercept to the y axis */

		sprintf (s, "x=%6.2f ", x);

		for (j=0; j<Datasets; j++)
		{
			y = getdata (x,j,cxdata,cydata,Datapoints,Datasets);
			sprintf (t, " y%d=%6.2f ",j+1,y);
			strcat (s,t);

			color.red=ltcolortable[j].red<<8; color.green=ltcolortable[j].green<<8; color.blue=ltcolortable[j].blue<<8; 
			/*dp (2, "Clickhandler: Setting color %02X%02X%02X\n",ltcolortable[j].red,ltcolortable[j].green,ltcolortable[j].blue);
			gdk_gc_set_foreground (gc, &color);
			gdk_gc_set_background (gc, &color);*/

			gdk_draw_line (graphdsp_drawingarea->window, gc, UserX(x),UserY(y),UserX(rxmin),UserY(y));


		}

		
		g_statusbar_display (s);
	}
	gdk_gc_destroy (gc);


}





/* Ctrl-Shift-Clicked on the drawing area -> display function FWHM if present (DEBUG PORPOSES ONLY) */

void g_ctlsftclickhandler (int fx, int fy)
{
char s[255], t[255];
double x,y, max, fwhm;
int i, j, maxidx;
GdkGC *gc;
GdkColor ltgray = {0x9600,0x9600,0x9600};
GdkColor color;

	drawall (graphdsp_drawingarea);  /* Erase previous lines */

	max=cydata[0]; maxidx=0;
	for (i=1; i<Datapoints; i++)
	{
		if (cydata[i] > max)
		{
			max = cydata[i]; maxidx=i;
		}
	}

	fwhm = find_fwhm (cydata, Datapoints);

	/* Display a thin gray dashed line at the x value */

	gc = gdk_gc_new (graphdsp_drawingarea->window);	/* allocate mem & assign null values */
	gdk_gc_copy (gc, graphdsp_drawingarea->style->black_gc); /* Copy some values */
	gdk_gc_set_dashes (gc, 0, "\3\3", 2);
	gdk_gc_set_line_attributes (gc,1,GDK_LINE_ON_OFF_DASH,GDK_CAP_NOT_LAST,GDK_JOIN_MITER);

	if ((maxidx>=rxmin) && (maxidx<=rxmax))
	{
		gdk_draw_line (graphdsp_drawingarea->window, gc, UserX(maxidx),UserY(rymin),UserX(maxidx),UserY(rymax));
		gdk_draw_line (graphdsp_drawingarea->window, gc, UserX(maxidx-fwhm/2),UserY(max/2),UserX(maxidx+fwhm/2),UserY(max/2));

		sprintf (s, "Peak at x=%4d, FWHM is %4.1f ", maxidx, fwhm);

		g_statusbar_display (s);
	}
	gdk_gc_destroy (gc);


}





/***************************************************************

	Graphdsp window handler and GUI
	
****************************************************************/


gboolean on_graphdsp_drawingarea_configure_event
                                        (GtkWidget  *widget,
                                        GdkEventConfigure *event,
                                        gpointer drwgarea)
{
	drawall (drwgarea);

	return FALSE;
}


gboolean on_graphdsp_drawingarea_button_press_event
                                        (GtkWidget  *widget,
                                        GdkEventButton *event,
                                        gpointer user_data)
{
int x,y;
GdkModifierType state;

	gdk_window_get_pointer (widget->window, &x, &y, &state);
	dp (2, "*click* at (%d,%d), state is %04x\n",x,y,state);
	
	gmotion_lastx=x; gmotion_lasty=y;
	
	if (state & GDK_BUTTON1_MASK)
	{
		gdrag_x_beg = x; gdrag_y_beg = y;
		gbutton1down=1;
	}
	if (state & GDK_BUTTON2_MASK)
	{
		gr_drag_x_beg = x; gr_drag_y_beg = y;
		gbutton2down=1;
	}
	if (state & GDK_BUTTON3_MASK)
	{
		gm_drag_x_beg = x; gm_drag_y_beg = y;
		gbutton3down=1;
	}
	
  	return FALSE;


}

 
gboolean on_graphdsp_drawingarea_button_release_event
                                        (GtkWidget  *widget,
                                        GdkEventButton *event,
                                        gpointer user_data)
{
int x,y;
GdkModifierType state;

	gdk_window_get_pointer (widget->window, &x, &y, &state);

	/* Button_released cannot differentiate between buttons */

	gdrag_x_end = x; gdrag_y_end = y;

   	
	if ((gdrag_x_beg==x) && (gdrag_y_beg==y) && (gbutton1down))
	{
		if ((state & GDK_SHIFT_MASK) && (state & GDK_CONTROL_MASK)) 	/* Ctrl-Shift-click, DELETEME */
			g_ctlsftclickhandler (x,y);
		else if (state & GDK_SHIFT_MASK)				/* Shift-click, jump to nearby maximum */
			g_sftclickhandler (x,y);
		else
			g_clickhandler (x,y);					/* Simple click shows data values */
	}
	else if (gbutton1down)	/* Otherwise, end of dragging a ROI */
	{
		
	}
	
	if (gbutton3down) 
	{

	}

	/* release all button flags */
	gbutton1down=gbutton2down=gbutton3down=0;



  return FALSE;
}


gboolean on_graphdsp_drawingarea_expose_event (GtkWidget *widget,
                                        GdkEventExpose *event,
                                        gpointer drwgarea)
{

  	drawall (drwgarea);
	return FALSE;
}


void on_graphdsp_save_clicked (GtkButton *button, gpointer thiswindow)
{
gchar* filename;
FILE* FP;
int i,j;

	ChooseFile ("Save data as (XY tab delimited)", window1, 1, &filename);
	if (filename)
	{
		FP = fopen (filename, "rt");
		if (FP)	/* file exists */
		{
			fclose (FP);
			statusbar_display ("File exists. No data written.");
		}
		else
		{
			FP = fopen (filename, "wt");
			for (i=0; i<Datapoints; i++)
			{
				fprintf (FP, "%10f \t ", cxdata[i]);
				for (j=0; j<Datasets; j++)
				{
					fprintf (FP, "%10f \t ", cydata[i + Datapoints*j]);
				}
				fprintf (FP, "\n");
			}
	

			fclose (FP);
			statusbar_display ("X-Y ASCII file written");
		}
		free (filename);
	}

}


void on_graphdsp_saveimg_clicked (GtkButton *button, gpointer thiswindow)
{
gchar* filename;
gchar *fmt;
gchar msg[512];
FILE* FP;
int i,j;
GdkPixbuf *pixbuf;
gint success;
GError *error = NULL;


	ChooseFile ("Save graph as png image", window1, 1, &filename);
	if (filename)
	{

		fmt = strrchr(filename, '.');
		if (fmt)
			fmt++;
		else
			strcpy (fmt, "png");			/* Defaults to PNG */

		pixbuf = gdk_pixbuf_get_from_drawable (NULL, GDK_DRAWABLE(graphdsp_drawingarea),
											     gdk_colormap_get_system (),
											     0,0,0,0, 
											     graphdsp_drawingarea->allocation.width, 
											     graphdsp_drawingarea->allocation.height);
			
		success = gdk_pixbuf_save (pixbuf, filename, fmt, &error, NULL);
		if (success)
		{
			sprintf (msg, "%s image file written", fmt);
			statusbar_display (msg);
			return;
		}

		if (error)
		{
			dp (1, "%s\n", error->message);
			sprintf (msg, "Error: %s", error->message);
			return;
		}
		g_free (filename);
	}

}



void on_graphdsp_cancel_clicked (GtkButton *button, gpointer thiswindow)
{

	if (cxdata) { free (cxdata); cxdata=NULL; }
	if (cydata) { free (cydata); cydata=NULL; }

	graphwindow_exists = 0;
	gtk_widget_destroy (thiswindow);

}



void g_statusbar_display (const char* text)
{

guint context;

	context = gtk_statusbar_get_context_id((GtkStatusbar*)gstatusbar1, "gcontext");

	if (g_statusbar_was_used) /* pop the last message */
	{
		gtk_statusbar_pop((GtkStatusbar*)gstatusbar1, context);
		g_statusbar_was_used=0;
	}
	
	gtk_statusbar_push((GtkStatusbar*)gstatusbar1, context, text);
	g_statusbar_was_used = 1;
}


void g_window_set_title (const char *s)
{

	if (graphwindow_exists)
		gtk_window_set_title (GTK_WINDOW (graphwindow), s);

}


void create_graphwindow (void)
{
  GtkWidget *graphdsp_vbox1;
  GtkWidget *graphdsp_hbox1;
  GtkWidget *graphdsp_saveimg;
  GtkWidget *graphdsp_savetxt;
  GtkWidget *graphdsp_cancel;

  graphwindow = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_object_set_data (GTK_OBJECT (graphwindow), "graphwindow", graphwindow);
  gtk_window_set_title (GTK_WINDOW (graphwindow), "Graph Display");

  graphdsp_vbox1 = gtk_vbox_new (FALSE, 0);
  gtk_widget_ref (graphdsp_vbox1);
  gtk_object_set_data_full (GTK_OBJECT (graphwindow), "graphdsp_vbox1", graphdsp_vbox1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (graphdsp_vbox1);
  gtk_container_add (GTK_CONTAINER (graphwindow), graphdsp_vbox1);

  graphdsp_drawingarea = gtk_drawing_area_new ();
  gtk_widget_ref (graphdsp_drawingarea);
  gtk_object_set_data_full (GTK_OBJECT (graphwindow), "graphdsp_drawingarea", graphdsp_drawingarea,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (graphdsp_drawingarea);
  gtk_box_pack_start (GTK_BOX (graphdsp_vbox1), graphdsp_drawingarea, TRUE, TRUE, 0);
  gtk_widget_set_usize (graphdsp_drawingarea, 320, 200);

  graphdsp_hbox1 = gtk_hbox_new (FALSE, 0);
  gtk_widget_ref (graphdsp_hbox1);
  gtk_object_set_data_full (GTK_OBJECT (graphwindow), "graphdsp_hbox1", graphdsp_hbox1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (graphdsp_hbox1);
  gtk_box_pack_start (GTK_BOX (graphdsp_vbox1), graphdsp_hbox1, FALSE, TRUE, 0);

  graphdsp_savetxt = gtk_button_new_with_label ("Save Tabular Data");
  gtk_widget_ref (graphdsp_savetxt);
  gtk_object_set_data_full (GTK_OBJECT (graphwindow), "graphdsp_savetxt", graphdsp_savetxt,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (graphdsp_savetxt);
  gtk_box_pack_start (GTK_BOX (graphdsp_hbox1), graphdsp_savetxt, TRUE, TRUE, 8);

//   graphdsp_saveimg = gtk_button_new_with_label ("Save As Image");
//   gtk_widget_ref (graphdsp_saveimg);
//   gtk_object_set_data_full (GTK_OBJECT (graphwindow), "graphdsp_saveimg", graphdsp_saveimg,
//                             (GtkDestroyNotify) gtk_widget_unref);
//   gtk_widget_show (graphdsp_saveimg);
//   gtk_box_pack_start (GTK_BOX (graphdsp_hbox1), graphdsp_saveimg, TRUE, TRUE, 8);

  graphdsp_cancel = gtk_button_new_with_label ("Dismiss");
  gtk_widget_ref (graphdsp_cancel);
  gtk_object_set_data_full (GTK_OBJECT (graphwindow), "graphdsp_cancel", graphdsp_cancel,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (graphdsp_cancel);
  gtk_box_pack_start (GTK_BOX (graphdsp_hbox1), graphdsp_cancel, TRUE, TRUE, 0);

  gstatusbar1 = gtk_statusbar_new ();
  gtk_widget_ref (gstatusbar1);
  gtk_object_set_data_full (GTK_OBJECT (graphwindow), "gstatusbar1", gstatusbar1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (gstatusbar1);
  gtk_box_pack_start (GTK_BOX (graphdsp_vbox1), gstatusbar1, FALSE, FALSE, 0);


	/* This is not handled by Glade: Under GTK+, low-level events are
	   not sent to the application by default (Harlow, p. 313).
	   We must specifically enable the mouse events */
	   
	gtk_widget_set_events (graphdsp_drawingarea, GDK_EXPOSURE_MASK
										| GDK_ENTER_NOTIFY_MASK
										| GDK_LEAVE_NOTIFY_MASK
										| GDK_BUTTON_PRESS_MASK
										| GDK_BUTTON_RELEASE_MASK
										| GDK_POINTER_MOTION_MASK
										| GDK_POINTER_MOTION_HINT_MASK
										| GDK_KEY_PRESS_MASK
										/* | GDK_ALL_EVENTS_MASK */);




	gtk_signal_connect (GTK_OBJECT (graphdsp_drawingarea), "configure_event",
                    	GTK_SIGNAL_FUNC (on_graphdsp_drawingarea_configure_event), graphdsp_drawingarea);
	gtk_signal_connect (GTK_OBJECT (graphdsp_drawingarea), "button_press_event",
                    	GTK_SIGNAL_FUNC (on_graphdsp_drawingarea_button_press_event), graphdsp_drawingarea);
	gtk_signal_connect (GTK_OBJECT (graphdsp_drawingarea), "button_release_event",
                    	GTK_SIGNAL_FUNC (on_graphdsp_drawingarea_button_release_event), graphdsp_drawingarea);
	gtk_signal_connect (GTK_OBJECT (graphdsp_drawingarea), "expose_event",
                    	GTK_SIGNAL_FUNC (on_graphdsp_drawingarea_expose_event), graphdsp_drawingarea);
	gtk_signal_connect (GTK_OBJECT (graphdsp_savetxt), "clicked",
                    	GTK_SIGNAL_FUNC (on_graphdsp_save_clicked), graphwindow);
/*	gtk_signal_connect (GTK_OBJECT (graphdsp_saveimg), "clicked",
                    	GTK_SIGNAL_FUNC (on_graphdsp_saveimg_clicked), graphwindow);*/
	gtk_signal_connect (GTK_OBJECT (graphdsp_cancel), "clicked",
                    	GTK_SIGNAL_FUNC (on_graphdsp_cancel_clicked), graphwindow);
	gtk_signal_connect (GTK_OBJECT (graphwindow), "destroy",
                    	GTK_SIGNAL_FUNC (on_graphdsp_cancel_clicked), graphwindow);

	graphwindow_exists = 1;
    gtk_widget_show (graphwindow);
  
}



void graphme (float* xdata, float* ydata, int numpts)
{

	/* Initially, the pointers are NULL. If any prior call has allocated memory,
	    we have to discard it in case the new numpts is higher than the old one */

	if (cxdata) free (cxdata);
	if (cydata) free (cydata);
	cxdata = (float*) calloc (numpts, sizeof (float));
	cydata = (float*) calloc (numpts, sizeof (float));
	if (!cxdata || !cydata) fatal_memerr();
	
	/* This allows the calling program to discard the data once it is displayed */
	
	memcpy (cxdata, xdata, numpts*sizeof (float));
	memcpy (cydata, ydata, numpts*sizeof (float));
	Datapoints = numpts; Datasets=1;

	if (graphwindow_exists)
		drawall (graphdsp_drawingarea);		
	else
		create_graphwindow ();
	
	
}

void multi_graphme (float* xdata, float* ydata, int numpts, int datasets)
{

	/* Initially, the pointers are NULL. If any prior call has allocated memory,
	    we have to discard it in case the new numpts is higher than the old one */

	if (cxdata) free (cxdata);
	if (cydata) free (cydata);
	cxdata = (float*) calloc (numpts, sizeof (float));
	cydata = (float*) calloc (numpts*datasets, sizeof (float));
	if (!cxdata || !cydata) fatal_memerr();
	
	/* This allows the calling program to discard the data once it is displayed */
	
	memcpy (cxdata, xdata, numpts*sizeof (float));
	memcpy (cydata, ydata, numpts*datasets*sizeof (float));
	Datapoints = numpts; Datasets=datasets;

	if (graphwindow_exists)
		drawall (graphdsp_drawingarea);		
	else
		create_graphwindow ();
	
	
}



void graphme_d (double* xdata, double* ydata, int numpts)
{
int i;

	/* Initially, the pointers are NULL. If any prior call has allocated memory,
	    we have to discard it in case the new numpts is higher than the old one */

	if (cxdata) free (cxdata);
	if (cydata) free (cydata);
	cxdata = (float*) calloc (numpts, sizeof (float));
	cydata = (float*) calloc (numpts, sizeof (float));
	if (!cxdata || !cydata) fatal_memerr();
	
	/* This allows the calling program to discard the data once it is displayed */
	
	for (i=0; i<numpts; i++)
	{
		cxdata[i] = xdata[i];
		cydata[i] = ydata[i];
	}
	Datapoints = numpts; Datasets=1;

	if (graphwindow_exists)
		drawall (graphdsp_drawingarea);		
	else
		create_graphwindow ();
	
	
}

void multi_graphme_d (double* xdata, double* ydata, int numpts, int datasets)
{
int i,j; 

	/* Initially, the pointers are NULL. If any prior call has allocated memory,
	    we have to discard it in case the new numpts is higher than the old one */

	if (cxdata) free (cxdata);
	if (cydata) free (cydata);
	cxdata = (float*) calloc (numpts, sizeof (float));
	cydata = (float*) calloc (numpts*datasets, sizeof (float));
	if (!cxdata || !cydata) fatal_memerr();
	
	/* This allows the calling program to discard the data once it is displayed */
	
	for (i=0; i<numpts; i++)
		cxdata[i] = xdata[i];
	for (j=0; j<datasets; j++)
		for (i=0; i<numpts; i++)
			cydata[i+numpts*j] = ydata[i+numpts*j];
			
	Datapoints = numpts; Datasets=datasets;

	if (graphwindow_exists)
		drawall (graphdsp_drawingarea);		
	else
		create_graphwindow ();
	
	
}


