

/*
 *	Segment2.c: More algorithms for segmentation
 * (Less-developed and untested algorithms)
 */



/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2012 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "segment.h"
#include "graphdsp.h"
#include "gtkhelpers.h"
#include "imgbuf.h"
#include "dlgbox.h"
#include "application.h"
#include "secondwnd.h"
#include "roi.h"
#include "functions.h"
#include "functions2.h"
#include "special1.h"
#include "filter.h"
#include "findclust.h"

















/**************************************************************************************************/

/*          ***    ****     ***    *   *           *   *   *****   ****     ***    *****          */
/*         *   *   *   *   *   *   *   *           ** **   *       *   *   *   *   *              */
/*         *       *   *   *   *   *   *           * * *   *       *   *   *       *              */
/*         *  **   ****    *   *   * * *   *****   * * *   ****    ****    *  **   ****           */
/*         *   *   * *     *   *   * * *           *   *   *       * *     *   *   *              */
/*         *   *   *  *    *   *   * * *           *   *   *       *  *    *   *   *              */
/*          ***    *   *    ***     * *            *   *   *****   *   *    ***    *****          */
/*                                                                                                */
/**************************************************************************************************

	2D grow & merge. This is a modification of the 
	region growing algorithm where multiple regions
	are grown. They merge if their average value
	falls within a threshold range.

	NOTE: This has never been fully tested and may well be incomplete

******************************************************/


#define BINS	50

/* rg_similarity checks for similar regions. Region's stats are defined by
   its average and sdev. Regions are similar if the lower avg plus its SD
   is higher than the higher avg minus its sd.
   Other similarity criteria may be formulated, for example a more rigid
   version where the averages _and_ the SDs have to match independently 
	Increase the variable 'tolerance' and regions are more likely to merge */


int rg_similarity (float tolerance, float Sx1, float Sxx1, int N1, float Sx2, float Sxx2, int N2)
{
float avg1, avg2, sd1, sd2;

	avg1 = Sx1/N1; avg2 = Sx2/N2;
	if (avg1 > avg2)				/* Make sure avg1 is the smaller one */
	{
		avg1 = Sx2/N2; avg2 = Sx1/N1;
		sd1 = SDEV(Sx2,Sxx2,N2); sd2 = SDEV(Sx1,Sxx1,N1);
	}
	else
	{
		sd1 = SDEV(Sx1,Sxx1,N1); sd2 = SDEV(Sx2,Sxx2,N2);
	}

	if (avg1+sd1+tolerance >= avg2-sd2) return 1;		/* Similar */

	return 0;								/* not similar */
}


/* SEED POINT DETERMINATION:
		Several strategies are possible (and intended). Presently, only one strategy is implemented.
		We compute the source image's histogram and check for peaks. In the seed pass, 
		pixels that fall within hte histogram bin of the histogram maximum are set in a fashion
		that connected clusters with the same value (actually, the same bin value range)
		become one seed cluster with a unique number. The purpose of the seed functions is to
		1) enter all seed points into a mask, and
		2) return the number of seed points (or rather seed clusters)

		The present implementation is hopelessly inefficient, but can serve as a starting point
		for further improvements. Note that flood-filling may be a good replacement in this case.

		Also note that grow-merge wants a pre-segmented image.

*/


int growmerge_fill_seed (image_type* img, image_type *mask, int xs, int ys, int z, int n, float binval, float delta)
{
short* p;
long idx;
int itercnt, changecnt, total_set, allow_grow;
int x,y, bx1, bx2, by1, by2;

	g_assert (mask->imgtype==B_SHORT);
	p = mask->data;
	p[xs + img->xmax*ys]=n;					/* Set the initial seed point */

	bx1=MAX(0,xs-2); bx2=MIN(mask->xmax-1, xs+2);
	by1=MAX(0,ys-2); by2=MIN(mask->ymax-1, ys+2);

	itercnt=0; total_set=0;
	do
	{
		itercnt++; changecnt=0;
		dp (3, "        Iteration %d: ", itercnt);

		for (y=by1; y<=by2; y++)
			for (x=bx1; x<=bx2; x++)
			{
				idx = x + img->xmax*y;
				if (p[idx]==0)											/* Check if unvisited */
				{
					allow_grow=0;

					if (((x>0) && (p[idx-1]==n))						/* Check for 4-connected neighbor */
						|| ((x<img->xmax-1) && (p[idx+1]==n))
						|| ((y>0) && (p[idx-img->xmax]==n))
						|| ((y<img->ymax-1) && (p[idx+img->xmax]==n)))
					{
						allow_grow=1;
					}

					if (((x>0) && (p[idx-1]) && (p[idx-1]!=n))						/* Check for 4-connected neighbor that is not n */
						|| ((x<img->xmax-1) && (p[idx+1]) && (p[idx+1]!=n))
						|| ((y>0) && (p[idx-img->xmax]) && (p[idx-img->xmax]!=n))
						|| ((y<img->ymax-1) && (p[idx+img->xmax]) && (p[idx+img->xmax]!=n)))
					{
						allow_grow=0;			/* In that case, we aren't allowed to grow, supersedes the previous. */
					}


					if (allow_grow && (fabs(binval-readbuf_flt(img,x,y,z)) < delta))	/* Check for membership in bin */
					{
						p[idx] = n;									/* Only then add to present cluster */
						if (x<bx1+2) bx1=MAX(0,x-2);
						if (x>bx2-2) bx2=MIN(mask->xmax-1,x+2);
						if (y<by1+2) by1=MAX(0,y-2);
						if (y>by2-2) by2=MIN(mask->ymax-1,y+2);
						changecnt++; total_set++;
					}
				}
			}

		dp (3, " ... %d pixels added\n",changecnt);
	}
	while (changecnt>0);
	return total_set;
}

/* Growmerge_seed sets the seed points in mask. */


int growmerge_seed (image_type *img, image_type *mask, int zslice)
{
short *p;	/* for the mask image */
int i,j, x,y,bin,n;
int x1, x2, y1, y2;
float buf,min,max,delta;
float histo[BINS+5];					/* 50-bin histogram */
float xdata[BINS+5],ydata[BINS+5];

	p = mask->data;

	roi_bbox (img, &x1,&y1,&x2,&y2);	/* If a ROI is set, stay within ROI */
	imgminmax (img, &min, &max);		/* Get image min and max values */
	
	/* In the first step, we need to find seed points. We get those form the 
	   peaks of a filtered histogram */
	
	for (i=0; i<BINS; i++) histo[i]=0;
	delta = (max-min)/BINS;

	for (y=y1; y<=y2; y++)	
		for (x=x1; x<=x2; x++)
		{
			if (in_roi (x,y))
			{
				buf = readbuf_flt (img, x,y,zslice);
				if (!F_ISZERO(buf))					/* zero is not part of any feature */
				{
					bin = (buf-min)/delta;
					histo[bin]+=1;
				}
			}
		}

	/* Now filter the histogram to allow identification of peaks */
	
	for (j=0; j<3; j++)
	{
		ydata[0] = histo[0];
		ydata[1] = 0.25*histo[0]+0.25*histo[2] + 0.5*histo[1];
		for (i=2; i<BINS; i++) ydata[i] = (histo[i-2]+2*histo[i-1]+3*histo[i]+2*histo[i+1]+histo[i+2])/9;
	}
	if (!macrorun)		/* Display the histogram if not within a macro */
	{
		for (i=2; i<BINS; i++) xdata[i]=min+i*delta;
		graphme (xdata, ydata, BINS);
	}

	/* For each identified peak, go over the mask and set the related points. Related means:
		We seek contiguous regions with values inside the bin of the histogram peak.
		Each of these regions gets assigned a unique number and serves as cluster of
		seed points.
	*/
	
	n=1;
	for (i=1; i<BINS-2; i++)
	{
		if (	(ydata[i] > ydata[i+1]) && (ydata[i] > ydata[i-1]) )	/* identified a peak */
		{
			buf = min+(i+0.5)*delta;		/* The corresponding image value */
			dp (1, "Grow-Merge: Histogram peak at %f has first cluster number %d\n",buf,n+1);
			
			for (y=y1; y<=y2; y++)	
				for (x=x1; x<=x2; x++)
				{
					if (in_roi (x,y) 											/* Seed must be witin ROI */
						&& (p[x + mask->xmax*y]==0)								/* ... and be zero, i.e., unvisited */
						&& ( !F_ISZERO(readbuf_flt (img,x,y,zslice)))			/* ... and lie on a non-background pixel */
						&& ( fabs(buf-readbuf_flt(img,x,y,zslice)) < delta))	/* ... and fall within the current bin */
					{
						j=growmerge_fill_seed (img, mask, x, y, zslice, n, buf, delta);		/* ... then set this point as seed point */
						dp (2, "Grow-Merge seed: Grown %d seed points for seed cluster %d from %d, %d\n",j,n,x,y);
						if (j>0) n++;
					}
				}
		}
	}

	return n;		/* Number of seed points */

}


void grow_merge (image_type* img, int zslice, int seedmeth, float tolerance, int clust_sep, int preview)
{
short *p;	/* for the mask image */
image_type mask;
int i,j,k,n,n0,n1, x,y,changecnt,itercnt,regions;
int x1, x2, y1, y2, xk,yk;
float f,buf,min,max;
float *rcnt, *rsx, *rsxx;
int neighbors[9];
float neigh_vals[9];


	if (allocate_image (&mask, B_SHORT, img->xmax, img->ymax, 1)==-1)
		fatal_memerr ();
		
	p = mask.data;	/* This is calloc'ed above, so it should be all-zero */
	mask.minval.imin=0;  mask.maxval.imax=255;

	roi_bbox (img, &x1,&y1,&x2,&y2);	/* If a ROI is set, stay within ROI */
	imgminmax (img, &min, &max);		/* Get image min and max values */

	n=growmerge_seed (img, &mask, zslice);

	/* If we got more than 32000 seed points, we are in trouble. Not only contains mask
		negative values (which later cause a segfault), but also this is an
		unreasonably high number. In that case, just STOP */

	if (n>32000)
	{
		ShowMessage ("Problem: Too many seed points.\nDid you pre-segment the image?");
		return;
	}
	if (preview)
	{
		make_minmax (&mask);
		secondary_display (&mask, 0, "Grow-Merge Preview");
		return;
	}

	/* Allocate the memory for the statistical arrays */
	
	rcnt = calloc (n+10,sizeof (float));
	rsx  = calloc (n+10,sizeof (float));
	rsxx = calloc (n+10,sizeof (float));
	if ((!rcnt) || (!rsx) || (!rsxx)) fatal_memerr ();

	dp (1, "Grow-merge: Starting with %d seed points.\n",n);

	regions=0;
	for (y=y1; y<=y2; y++)					/* Go over the mask again; this time use it to */
		for (x=x1; x<=x2; x++)				/* ...initialize the statistical arrays */
		{
			if ((i=p[x+mask.xmax*y])>0)			/* SIC! don't use == here */
			{
				rcnt[i]++;
				buf = readbuf_flt (img,x,y,zslice);
				rsx[i] += buf;
				rsxx[i] += SQR(buf);
				regions = MAX(i,regions);
			}
		}

	/* At this time, we have a number of regions, contiguously
	   numbered from 1 to 'regions'. For each region, we accumulate
	   statistics as we grow (mean, sdev). */


	itercnt=0;
	do
	{
		itercnt++; changecnt=0;
		if (!(itercnt & 0x07)) dp (2, "Iteration %d: ", itercnt);
	
		for (y=y1; y<=y2; y++)
			for (x=x1; x<=x2; x++)
			{
				/* Consider only points which are not yet set and which have
					one or more nonzero pixels in the mask neighborhood. Any zero-valued pixel
					with a nonzero neighbor is a candidate for growing */
			
				if (in_roi(x,y) && (p[x + mask.xmax*y]==0) && (F_ISZERO( (f=readbuf_flt(img,x,y,zslice) ))))
				{

					/* Find all nonzero neighbors, line'em up in 'neighbors' array */
					
					for (j=0; j<9; j++) neighbors[j]=0;
					k=0;
					for (i=MAX(0,y-1); i<=MIN(mask.ymax-1,y+1); i++)
						for (j=MAX(0,x-1); j<=MIN(mask.xmax-1,x+1); j++)
						{
							if (p[j+mask.xmax*i]) 	/* Note: The central pixel is always zero */
							{
								neighbors[k]=p[j+mask.xmax*i];
								neigh_vals[k] = readbuf_flt (img, j,i,zslice);
								k++;
							}
						}
					
					if (k>0)		/* means, we have at least one nonzero neighbor b/c the central pixel is zero */
					{
						/* Find the region with the gray value closest to the present pixel */
						
						buf = 10*(max-min+1); n=0;		/* Just a large number */
						for (j=0; j<k; j++)
						{
							if ( fabs(f-neigh_vals[j]) < buf)
							{
								buf = fabs(f-neigh_vals[j]);	/* Find the closest neighbor */
								n = j;
							}
						}

						/* Push the neighbor with the lowest grayval distance to position 0 */
						buf = neigh_vals[0]; neigh_vals[0]=neigh_vals[n]; neigh_vals[n]=buf;
						n0 = neighbors[0]; neighbors[0]=neighbors[n]; neighbors[n]=n0;
						
						n0=neighbors[0];
						buf = readbuf_flt (img, x,y,zslice);	/* buf=f? */
						p[x+mask.xmax*y] = n0;	/* Pixel is attached to the neighbor with the nearest grayvalue distance */
						changecnt++;			/* Count the number of added pixels */
						rcnt[n0]++;				/* Update statistics for this cluster */
						rsx[n0] += buf;
						rsxx[n0] += SQR(buf);

						for (j=1; j<k; j++)
							if (neighbors[j] != n0)		/* These are the two regions we can possibly merge */
							{
								n1 = neighbors[j];
								if (rg_similarity (tolerance, rsx[n1],rsxx[n1],rcnt[n1], rsx[n0],rsxx[n0],rcnt[n0]))
								{
									/* The regions numbered n0 and n1 are similar so
										sweep over the image and convert n1 into n0 */
									
									dp (3, "Grow-merge: Merging regions %d and %d\n",n0,n1);
									for (yk=y1; yk<=y2; yk++)
										for (xk=x1; xk<=x2; xk++)
											if (p[xk+mask.xmax*yk]==n1) p[xk+mask.xmax*yk]=n0;

									rcnt[n0] += rcnt[n1]; rsx[n0]+=rsx[n1]; rsxx[n0]+=rsxx[n1]; 
									rcnt[n1]=0; rsx[n1]=0; rsxx[n1]=0;		/* Region n1 no longer exists. remove */
								}

							}
					}

				}
			
			}
	
		if (!(itercnt & 0x07)) dp (2, "%d pixels\n",changecnt);
	}
	while (changecnt>0);


	/* If we requested cluster separation, delete boundary pixels */
	
	if (clust_sep)
	{
		for (y=y1; y<=y2; y++)
			for (x=x1; x<=x2; x++)
				if ((n=p[x+mask.xmax*y]))		/* Check if mask is set. Do not use == here */
				{
					n0=0;
					for (i=MAX(0,y-1); i<=MIN(mask.ymax-1,y+1); i++)
						for (j=MAX(0,x-1); j<=MIN(mask.xmax-1,x+1); j++)
						{
							if ((p[j+mask.xmax*i]) && (n!=p[j+mask.xmax*i])) n0=1;
						}
					if (n0) p[x+mask.xmax*y]=0;		/* Delete identified boundary pixel */
				}

	}


	/* What follows is a clean-up step where the regions are consecutively numbered */

	for (i=0; i<=regions; i++)
	{
		if (rcnt[i]) rsx[i] /= rcnt[i]; else rsx[i]=regions+10;
		rcnt[i]=i;						/* We only need the average, so discard the rest */
	}

	/* Sort regions by ascending average value */

	for (i=0; i<regions; i++)
	{
		buf=rsx[i]; n=i;
		for (j=i+1; j<=regions; j++)		/* Dirtysort. Inefficient, but... */
			if (rsx[j] < buf)
			{
				buf=rsx[j]; n=j;		/* Find minimum. */
			}
		
		if (i!=n)
		{
			buf=rsx[i]; rsx[i]=rsx[n]; rsx[n]=buf;
			buf=rsxx[i]; rsxx[i]=rsxx[n]; rsxx[n]=buf;
			j=rcnt[i]; rcnt[i]=rcnt[n]; rcnt[n]=j;
		}
	}

	/* Translate in mask image */

	for (y=y1; y<=y2; y++)
		for (x=x1; x<=x2; x++)
			if ((n=p[x+mask.xmax*y]))		/* Do not use == here */
			{
				j=0; while ((j<regions) && (rcnt[j]!=n)) j++;
				if (rcnt[j]==n) p[x+mask.xmax*y]=j;
			}



	free (rcnt); free (rsx); free (rsxx);

	if (preview)
	{
		make_minmax (&mask);
		secondary_display (&mask, 0, "Grow-Merge Preview");
	}
	else
		handback (img, &mask);

}




GtkWidget *gm_window;
GtkAdjustment* gm_tolerance_adj;
GtkWidget *gm_preview;
GtkWidget *gm_bound;


void  on_gm_preview_toggled (GtkWidget *nothing, gpointer user_data)
{


}

void  on_gm_bound_toggled (GtkWidget *nothing, gpointer user_data)
{


}

void on_gm_ok_clicked ( GtkButton * button, gpointer data )
{
int do_preview, do_bound;
float tolerance;
image_type* img;

	img = data;
	undo_nexus (img);

	do_preview = gtk_toggle_button_get_active ((GtkToggleButton*) gm_preview);
	do_bound = gtk_toggle_button_get_active ((GtkToggleButton*) gm_bound);
	tolerance = gm_tolerance_adj->value;
	
	grow_merge (img, cur_slice, 0, tolerance, do_bound, do_preview);

	if (img == &mainimg) 
	{
		make_imlib_buffer_from_main_image (cur_slice);
		redraw (1);
	}

	gtk_widget_destroy (gm_window);
}


void on_gm_cancel_clicked ( GtkButton * button, gpointer data )
{

	gtk_widget_destroy (gm_window);
}




void create_gm_window (image_type* img)
{
  GtkWidget* vbox;
  GtkWidget *gm_table1;
  GtkWidget *gm_label1;
  GtkWidget *gm_label2;
  GtkWidget *gm_label3;
  GtkWidget *gm_tolerance;
  GtkWidget *gm_okbutton;
  GtkWidget *gm_cancelbutton;
  
  float min, max;
  
	imgminmax (img, &min, &max);

	gm_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (gm_window), "gm_window", gm_window);
	gtk_container_set_border_width (GTK_CONTAINER (gm_window), 3);
	gtk_window_set_title (GTK_WINDOW (gm_window), "Grow-Merge");
	vbox = vbox_in_window (gm_window, "gm_vbox");
	gm_table1 = MakeTable (gm_window, vbox, "gm_table1", 2,4);

	gm_label1 = MakeLabel (gm_window,gm_table1,"Tolerance","gm_label1",0,0);
	gm_label2 = MakeLabel (gm_window,gm_table1," ","gm_label4",0,1);
	gm_label3 = MakeLabel (gm_window,gm_table1," ","gm_label5",0,2);

	gm_tolerance_adj = GTK_ADJUSTMENT (gtk_adjustment_new (0, -500, 500, 1, 10, 0));
	gm_tolerance = MakeHScale (gm_tolerance_adj,gm_window,gm_table1,"gm_lthresh",1,0);
	gtk_scale_set_digits (GTK_SCALE (gm_tolerance), 0);

	gm_preview = Checkbutton_in_table (gm_window,gm_table1,"Preview","gm_prv",1,1,on_gm_preview_toggled);

	gm_bound = Checkbutton_in_table (gm_window, gm_table1, "Separate clusters", "gm_toggle", 1,2,on_gm_bound_toggled);

	gm_okbutton = Button_in_table(gm_window,gm_table1,"OK","gm_okbutton",0,4,NULL);
	gm_cancelbutton = Button_in_table(gm_window,gm_table1,"Cancel","gm_cancelbutton",1,4,NULL);

	gtk_signal_connect (GTK_OBJECT (gm_window), "destroy",
                      GTK_SIGNAL_FUNC (on_gm_cancel_clicked), gm_window);
	gtk_signal_connect (GTK_OBJECT (gm_okbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_gm_ok_clicked),img);
	gtk_signal_connect (GTK_OBJECT (gm_cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_gm_cancel_clicked),gm_window);

	gtk_widget_show (gm_window);
}


void grow_merge_main ()
{
	create_gm_window (&mainimg);
}


/*********************************************************************************************************

 #   #    #+#    #       #                ##*    #        #*#    #   #   ####     #*#    #   #    ##*   
 #   #     #     #       #               #   #   #         #     ## ##   #   #     #     #   #   #   #  
 #   #     #     #       #               #       #         #     # # #   #   #     #     ##  #   #      
 #####     #     #       #       #####   #       #         #     # # #   ####      #     # # #   #  ##  
 #   #     #     #       #               #       #         #     #   #   #   #     #     #  ##   #   #  
 #   #     #     #       #               #   #   #         #     #   #   #   #     #     #   #   #   #  
 #   #    #*#    #####   #####            ##+    #####    #*#    #   #   ####     #*#    #   #    ##*   

**********************************************************************************************************/
 
/* Hill climbing was originally published by Bankmann et al, IEEE Trans Inform Techn Biomed 1997.

	Hill-climbing is another variant of region growing with automated seeding.
	The seed points are the local maxima of the image, and in some analogy to the
	Spider segmentation algorithm (Haidekker et al, BJR 1997), a polygonal approximation
	of the boundary is sought. From the boundary points, region growing takes place
	with the constraints
	-	Only uphill points can be grown (i.e., intensity equal or higher than boundary point)
	-	Growing takes place only in the direction of the seed point
	Note that I added the maximum distance to the seed point as one additional constraint.
	Note also that this method becomes much more reliable when the source image is heavily blurred.
*/



/* Return the normalized slope for local maximum at (x0,y0) and point (x,y) */

float hillc_slope (image_type *img, int x0, int y0, int x, int y)
{
float buf1, buf2, slope;

	if ((x0==x) && (y0==y)) return 0.0;
	buf1 = readbuf_flt (img, x0, y0, 0);
	buf2 = readbuf_flt (img, x, y, 0);
	slope = (buf1 - buf2) / sqrt(ARITHM( (x0-x), (y0-y) ));
	return slope;
}


/* Wnter a colored pixel into RGB image vizimg */

void hillc_label (image_type *vizimg, int x, int y, int color)
{
rgbtriplet rgb;

	switch (color)
	{
		case 0:		rgb = (rgbtriplet) {0,0,0}; break;					/* Black */
		case 1:		rgb = (rgbtriplet) {0, 0x60, 0xff}; break;			/* Light blue */
		case 2:		rgb = (rgbtriplet) {0xff, 0xa0, 0x20}; break;		/* Orangeish-yellow */
		case 3:		rgb = (rgbtriplet) {0x10, 0xff, 0x00}; break;		/* Green */
	}
	writebuf (vizimg, &rgb, x,y,0);
}

/* Depending on the direction from the cusp, set the growth pattern. The pattern
	contains 1 for pixels that are allowed to grow, 0 otherwise. The bits are packed
	into a byte with the following pattern (bit numbers around central pixel marked * )

		7  0  1
		6  *  2
		5  4  3
*/

int get_dirmask (int x, int y)
{
float theta;

	theta = atan2 (y,x)*180.0/PI;							/* Easier in degrees */

	if ((theta < 22.5) && (theta > -22.5))					/* North of seed pixel */
		return 0x7c;										/* 0111-1100 */
	else if ((theta >= 22.5) && (theta < 67.5))				/* Northeast of seed pixel */
		return 0xf8;										/* 1111-1000 */
	else if ((theta >= 67.5) && (theta < 112.5))			/* East of seed pixel */
		return 0xf1;										/* 1111-0001 */
	else if ((theta >= 112.5) && (theta < 157.5))			/* Southheast of seed pixel */
		return 0xe3;										/* 1110-0011 */
	else if ((theta <= -22.5) && (theta > -67.5))			/* Northwest of seed pixel */
		return 0x3e;										/* 0011-1110 */
	else if ((theta <= -67.5) && (theta > -112.5))			/* West of seed pixel */
		return 0x1f;										/* 0001-1111 */
	else if ((theta <= -112.5) && (theta > -157.5))			/* Southwest of seed pixel */
		return 0x8f;										/* 1000-1111 */
	else 													/* Last case left: South */
		return 0xc7;										/* 1100-0111 */
}



/* The core algorithm performs hillclimbing on ONE hill with cusp at (xs,ys).
	Makes heavy use of the Bresenham algorithm for straight lines */

void hillclimb_core (image_type *img, image_type *mask, image_type *vizimg, int xs, int ys, int numrays, int feature_radius)
{
int x0,y0,x1,y1;							/* Start and target for the line */
int dx, dy, ax, ay, sgx, sgy, d, x, y;		/* Bresenham variables */
float theta;								/* Angular increment */
int ray, in_range, mx, my, i,j,k, cnt, itercnt;
int rmax;
float slope, max_slope;
float buf1, buf2;
vertex peripts[numrays+1];					/* Points that define the perimeter */
int dirmask;
unsigned char allowed_dirs[9] = {0x80, 0x01, 0x02, 0x40, 0x00, 0x04, 0x20, 0x10, 0x08};
/*								   NW    N     NE    W     c     E     SW    S     SE  */


	theta = TWOPI/numrays;

	for (ray=0; ray<numrays; ray++)
	{
		x1 = (int)(0.4999 + xs + feature_radius*cos(ray*theta));
		y1 = (int)(0.4999 + ys + feature_radius*sin(ray*theta));
		x0 = xs; y0 = ys; mx=x0; my=y0; max_slope=0.0;


		/* Bresenham line from (x0,y0) to (x1,y1) */

		dx=x1-x0; dy=y1-y0; ax=2*abs(dx); ay=2*abs(dy);
		sgx=isign(dx); sgy=isign(dy);
		x=x0; y=y0;
		if (ax>ay)        /* x dominant */
		{
			d = ay - dx;
			while (x!=x1)
			{
				if (d>=0) { y+=sgy; d-=ax; }
				x+=sgx; d+=ay;
				hillc_label (vizimg, x,y, 1);
				slope = hillc_slope (img, xs,ys, x,y);
				if (slope > max_slope)
				{
					max_slope = slope; mx=x; my=y;
				}
			}
		}
		else              /* y dominant */
		{
			d = ax - dy;
			while (y!=y1)
			{
				if (d>=0) { x+=sgx; d-=ay; }
				y+=sgy; d+=ax;
				hillc_label (vizimg, x,y, 1);
				slope = hillc_slope (img, xs,ys, x,y);
				if (slope > max_slope)
				{
					max_slope = slope; mx=x; my=y;
				}
			}
		}

		/* Line done. Maximum slope should be at (mx,my) */

		hillc_label (vizimg, mx,my, 2);
		peripts[ray].x = mx; peripts[ray].y = my;
	}

	/* End probing all rays. We have the peripts array ready, so start hill climbing. Since we know
		the size of the features, we can delimit a ROI. This ROI excludes the boundary strip. */

	x0 = xs; y0 = ys;
	x1 = xs; y1 = ys;
	rmax = 0;

	for (i=0; i<numrays; i++)
	{
		x = peripts[i].x; y = peripts[i].y;
		((unsigned char *)mask->data)[x + mask->xmax * y] = 1;		/* Seed points */
		x0 = MIN (x0, x); y0 = MIN (y0, y);
		x1 = MAX (x1, x); y1 = MAX (y1, y);
		rmax = MAX (rmax, SQR(x-xs)+SQR(y-ys));						/* Radius of circumscribed circle */
	}
	x0 = MAX (x0, 1);
	y0 = MAX (y0, 1);
	x1 = MIN (x1, mask->xmax-2);
	y1 = MIN (y1, mask->ymax-2);

	itercnt=0;
	do										/* Iterative region-growing */
	{
		cnt=0; itercnt++;
		for (y=y0; y<=y1; y++)
			for (x=x0; x<=x1; x++)
			{
				if (((unsigned char *)mask->data)[x + mask->xmax * y])		/* Point has been set, need to */
				{															/* examine its neighbors */
					dirmask = get_dirmask (x-xs, y-ys);						/* Allowable pixels for growth? */
					buf1 = readbuf_flt (img, x,y,0);

					/* Grow a pixel in the 3x3 neighborhood if dirmask allows it and if we go uphill */

					k=0;
					for (j=y-1; j<=y+1; j++)
						for (i=x-1; i<=x+1; i++)
						{
							if ((dirmask & allowed_dirs[k])
									&& (SQR(i-xs)+SQR(j-ys) <= rmax)
									&& (((unsigned char *)mask->data)[i + mask->xmax * j]==0))
							{
								buf2 = readbuf_flt (img, i,j,0);
								if (buf2 >= buf1)							/* Uphill! */
								{
									((unsigned char *)mask->data)[i + mask->xmax * j] = 1;		/* New seed point */
									hillc_label (vizimg,i,j,3);				/* Label it in control image */
									cnt++;
								}
							}
							k++;
						}
				}
			}
	}
	while ((cnt>0));						/* Iterate as long as new points have been added */

	/* In the last step (since we have the ROI at this point), copy the pixels to slice 2 of the mask
		and reset them in slice 1. */

	for (y=y0; y<=y1; y++)
		for (x=x0; x<=x1; x++)
		{
			if (((unsigned char *)mask->data)[x + mask->xmax * y])
			{
				((unsigned char *)mask->data)[x + mask->xmax * (y + mask->ymax)] = 1;
				((unsigned char *)mask->data)[x + mask->xmax * y] = 0;
			}
		}


}


/* The main driver for the hill climbing algorithm, as called from GUI and macro interpreter */

/* TODO: Thia assumes a 2D image and acts on slice 0 only */
/* TODO: RGB is -- as it is way too often -- reduced to luminance only */

void hillclimb_segmentation (image_type *img, float thresh, int locmax_radius, int numrays, int feature_radius)
{
image_type target, vizimg, mask;
int x,y,i, nummax;
float buf,min,max;
rgbtriplet rgb;
float ray_data[feature_radius];
unsigned char zerobuf[4] = {0,0,0,0};


	/* We begin by finding the local maxima in the target image 
		and preparing a RGB image for visualization */
	
	imgminmax (img, &min, &max);
	if (allocate_image (&target, img->imgtype, img->xmax, img->ymax, 1) == -1)
		fatal_memerr();
	if (allocate_image (&mask, B_SHORT, img->xmax, img->ymax, 2) == -1)
		fatal_memerr();
	if (allocate_image (&vizimg, B_RGB, img->xmax, img->ymax, 1) == -1)
		fatal_memerr();
	copybuf_z (img, &target, 0, 0);								/* Copy only slice 0 to target */
	nummax = locmax_uep (&target, thresh, locmax_radius, 0);		/* Actual 2D local maximum finder */

	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{
			buf = readbuf_flt (img,x,y,0);
			i = (int)(128.0*(buf-min)/(max-min));
			rgb.red = rgb.green = rgb.blue = i;			/* Darkened version of origial in grayscale */
			buf = readbuf_flt (&target,x,y,0);
			if (buf>0.0) rgb.red=255;					/* Local maxima are bright red */
			writebuf (&vizimg, &rgb, x,y,0);

		}


	/* Next step, trace the rays from each local maximum to the point of maximum slope */

	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{
			if (!F_ISZERO(readbuf_flt (&target,x,y,0)))		/* A local maximum exists at (x,y) */
				hillclimb_core (img, &mask, &vizimg, x,y, numrays, feature_radius);
		}


	/* Create another copy of the current slice... now, we zero out everything that
		is not in the mask. This is the actual segmentation. Note that the prior call
		to locmax_uep has changed target, so we need to re-allocate it. */

	freebuf (&target);
	if (allocate_image (&target, img->imgtype, img->xmax, img->ymax, 1) == -1)
		fatal_memerr();
	copybuf_z (img, &target, 0, 0);

	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{
			if ( (((unsigned char *)mask.data)[x + mask.xmax * (y + mask.ymax)])==0 )
				writebuf (&target, &zerobuf, x,y,0);
		}

	if (!macrorun)
	{
		make_minmax (&vizimg);
		secondary_display (&vizimg, 0, "Hill-Climbing Visualization");
	}

	handback (img, &target);

}





GtkWidget *hillc_window;
GtkAdjustment* hillc_radius_adj;
GtkAdjustment* hillc_locrad_adj;
GtkAdjustment* hillc_bdpts_adj;
GtkAdjustment* hillc_thresh_adj;



void on_hillc_ok_clicked ( GtkButton * button, gpointer data )
{
int locmax_radius, numrays, feature_radius;
float thresh;
image_type* img;


	img = data;
	if (img == &mainimg) 
	{
		undo_nexus (img);
	}

	thresh = hillc_thresh_adj->value;
	locmax_radius = hillc_locrad_adj->value;
	numrays = hillc_bdpts_adj->value;
	feature_radius = hillc_radius_adj->value;
	hillclimb_segmentation (img, thresh, locmax_radius, numrays, feature_radius);

	if (img == &mainimg) 
	{
		make_imlib_buffer_from_main_image (cur_slice);
		redraw (1);
	}

	gtk_widget_destroy (hillc_window);
}


void on_hillc_cancel_clicked ( GtkButton * button, gpointer data )
{

	gtk_widget_destroy (hillc_window);
}




void create_hillc_window (image_type* img)
{
  GtkWidget* vbox;
  GtkWidget *hillc_table1, *hillc_label1, *hillc_label2, *hillc_label3, *hillc_label4;
  GtkWidget *hillc_locrad, *hillc_bdpts, *hillc_radius, *hillc_thresh;
  GtkWidget *hillc_okbutton;
  GtkWidget *hillc_cancelbutton;
  GtkTooltips *tt_maxrad, *tt_locrad, *tt_bpts, *tt_thresh;
  
  float min, max;
  int y, maxsize;
  
	imgminmax (img, &min, &max);
	maxsize = MAX(img->xmax, img->ymax);

	hillc_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (hillc_window), "hillc_window", hillc_window);
	gtk_container_set_border_width (GTK_CONTAINER (hillc_window), 3);
	gtk_window_set_title (GTK_WINDOW (hillc_window), "Hill-Climbing Semgentation");
	vbox = vbox_in_window (hillc_window, "hillc_vbox");
	hillc_table1 = MakeTable (hillc_window, vbox, "hillc_table1", 2,5);


	y=0;
	hillc_label1 = MakeLabel (hillc_window,hillc_table1,"Local maximum radius","hillc_label1",0,y);
	hillc_locrad_adj = GTK_ADJUSTMENT(gtk_adjustment_new (15, 1, MAX(15,maxsize/10), 1, 1, 0));
	hillc_locrad = MakeSpinbutton (hillc_locrad_adj,hillc_window, hillc_table1,"hillc_locmax_adj",1,y);
	tt_locrad = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_locrad, hillc_locrad, "The radius of the environment in which a maximum\nmust be absolute to count as local maximum", NULL);

	y++;
	hillc_label4 = MakeLabel (hillc_window,hillc_table1,"Peak threshold","hillc_label4",0,y);
	hillc_thresh_adj = GTK_ADJUSTMENT(gtk_adjustment_new (min+1, min, max, 1, 1, 0));
	hillc_thresh = MakeSpinbutton (hillc_thresh_adj,hillc_window, hillc_table1,"hillc_thresh_adj",1,y);
	tt_thresh = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_thresh, hillc_thresh, "Minimum image value for a pixel\nto count as local maximum", NULL);
	
	y++;
	hillc_label2 = MakeLabel (hillc_window,hillc_table1,"Number of boundary points","hillc_label4",0,y);
	hillc_bdpts_adj = GTK_ADJUSTMENT(gtk_adjustment_new (16, 8, 64, 1, 1, 0));
	hillc_bdpts = MakeSpinbutton (hillc_bdpts_adj,hillc_window, hillc_table1,"hillc_bdpts_adj",1,y);
	tt_bpts = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_bpts, hillc_bdpts, "Number of probing rays on which\nthe maximum slope is determined", NULL);

	y++;
	hillc_label3 = MakeLabel (hillc_window,hillc_table1,"Maximum feature radius","hillc_label5",0,y);
	hillc_radius_adj = GTK_ADJUSTMENT(gtk_adjustment_new (30, 1, MAX(30,maxsize/4), 1, 1, 0));
	hillc_radius = MakeSpinbutton (hillc_radius_adj,hillc_window, hillc_table1,"hillc_radius_adj",1,y);
	tt_maxrad = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_maxrad, hillc_radius, "Maximum allowable radius of any feature", NULL);

	y++;
	hillc_okbutton = Button_in_table(hillc_window,hillc_table1,"OK","hillc_okbutton",0,y,NULL);
	hillc_cancelbutton = Button_in_table(hillc_window,hillc_table1,"Cancel","hillc_cancelbutton",1,y,NULL);

	gtk_signal_connect (GTK_OBJECT (hillc_window), "destroy",
                      GTK_SIGNAL_FUNC (on_hillc_cancel_clicked), hillc_window);
	gtk_signal_connect (GTK_OBJECT (hillc_okbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_hillc_ok_clicked),img);
	gtk_signal_connect (GTK_OBJECT (hillc_cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_hillc_cancel_clicked),hillc_window);

	gtk_widget_show (hillc_window);
}


void hillclimbing_main ()
{
	create_hillc_window (&mainimg);
}




/*
TODO: Here is an interesting algorithm, also called ISODATA. Seen at
http://fourier.eng.hmc.edu/e161/lectures/classification/node12.html



The Isodata Algorithm

Isodata stands for Iterative Self-Organizing Data Analysis Techniques. This is a more sophisticated algorithm which allows the number of clusters to be automatically adjusted during the iteration by merging similar clusters and splitting clusters with large standard deviations. We first define the following parameters:

   1. K = number of clusters desired;
   2. I = maximum number of iterations allowed;
   3. P = maximum number of pairs of cluster which can be merged;
   4. $\Theta_N$ = a threshold value for minimum number of samples in each cluster can have (used for discarding clusters);
   5. $\Theta_S$ = a threshold value for standard deviation (used for split operation);
   6. $\Theta_C$ = a threshold value for pairwise distances (used for merge operation).

The algorithm:

    * Step 1. Arbitrarily choose $k$ (not necessarily equal to K) initial cluster centers: $ {\bf m}_1,\;{\bf m}_2,....,{\bf m}_k $ from the data set $\{{\bf x}_i,\;\;i=1,2,...,N\}$.

    * Step 2. Assign each of the $N$ samples to the closest cluster center:

      \begin{displaymath}{\bf x} \in \omega_j\;\;\;\;if\;\;\;D_L({\bf x},{\bf m}_j) =max\;\{\;D_L({\bf x},{\bf m}_i),\;\;\;i=1,\cdots,k) \end{displaymath}

    * Step 3. Discard clusters with fewer than $\Theta_N$ members, i.e., if for any $j$, $ N_j < \Theta_N $, then discard $\omega_j$ and $k \leftarrow k-1$.
    * Step 4. Update each cluster center:

      \begin{displaymath}{\bf m}_j=\frac{1}{N_j} \sum_{{\bf x} \in \omega_j} {\bf x} \;\;\;\;\;(j=1,\cdots,k) \end{displaymath}

    * Step 5. Compute the average distance $D_j$ of samples in cluster $\omega_j$ from their corresponding cluster center:

      \begin{displaymath}D_j=\frac{1}{N_j} \sum_{{\bf x} \in \omega_j} D_L({\bf x},{\bf m}_j) \;\;\;\;\; (j=1,\cdots,k) \end{displaymath}

    * Step 6. Compute the overall average distance of the samples from their respective cluster centers:

      \begin{displaymath} D=\frac{1}{N} \sum_{j=1}^k N_j D_j \end{displaymath}

    * Step 7. If $k \leq K/2$ (too few clusters), go to Step 8; else if $k>2K$ (too many clusters), go to Step 11; else go to Step 14.

      (Steps 8 through 10 are for split operation, Steps 11 through 13 are for merge operation.)
    * Step 8. First step to split. Find the standard deviation vector ${\bf\sigma}_j=[\sigma_1^{(j)},\cdots,\sigma_n^{(j)}]^T$ for each cluster:

      \begin{displaymath}\sigma_i^{(j)} = \sqrt{\frac{1}{N_j} \sum_{{\bf x} \in \omega... ... (x_i-m_i^{(j)})^2}, \;\;\;\;\;(i=1,\cdots,n,\;\;j=1,\cdots,k) \end{displaymath}

      where $m_i^{(j)}$ is the ith component of ${\bf m}_j$ and $\sigma_i$ is the standard deviation of the samples in $\omega_j$ along the ith coordinate axis. $N_j$ is the number of samples in $\omega_j$.

    * Step 9. Find the maximum component of each $\Sigma_j$ and denote it by $\sigma_{max}^{(j)}$; Do this for all $j=1,\cdots,k$.
    * Step 10.

      If for any $\sigma_{max}^{(j)},\;\;\;\;(j=1,\cdots,k)$, all of the following are true
          o $\sigma_{max}^{(j)} > \Theta_S$,
          o $ D_j > D$,
          o $ N_j > 2\Theta_N$

      then split ${\bf m}_j$ into two new cluster centers ${\bf m}_j^+$ and ${\bf m}_j^-$ by adding $\pm \delta$ to the component of ${\bf m}_j$ corresponding to $\sigma_{max}^{(j)}$, where $\delta$ can be $\alpha\;\sigma_{max}^{(j)}$, for some $\alpha>0$. Then delete ${\bf m}_j$ and let $k \leftarrow k+1$. Goto Step 2

      else Go to Step 14.

    * Step 11. First step to merge. Compute the pairwise distances $D_{ij}$ between every two cluster centers:

      \begin{displaymath}D_{ij}=D_L({\bf m}_i,{\bf m}_j),\;\;\;\;\;\mbox{for all $i\neq j$} \end{displaymath}

      and arrange these $k(k-1)/2$ distances in ascending order.

    * Step 12. Find no more than $P$ smallest $D_{ij}$'s which are also smaller than $\Theta_C$ and keep them in ascending order:

      \begin{displaymath}D_{i_1 j_1} \leq D_{i_2 j_2} \leq \cdots \leq D_{i_P j_P} \end{displaymath}

    * Step 13. Perform pairwise merge: for $l=1,\cdots,P$, do the following:

      If neither of ${\bf m}_{i_l}$ nor ${\bf m}_{j_l}$ has been used in this iteration, then merge them to form a new center:

      \begin{displaymath}{\bf m}=\frac{1}{N_{i_l}+N_{j_l}} [ N_{i_l} {\bf m}_{i_l}+N_{j_l}{\bf m}_{j_l}]\end{displaymath}

      Delete ${\bf m}_{i_l}$ and ${\bf m}_{j_l}$, and let $k \leftarrow k-1$.

      Go to Step 2.
    * Step 14. Terminate if maximum number of iterations I is reached. Otherwise go to Step 2.

The Isodata algorithm is more flexible than the K-mean method. But the user has to choose empirically many more parameters listed previously. 

*/

